# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref dataimport utilities for EAC-CPF (Encoded Archival
Context -- Corporate Bodies, Persons, and Families).
"""

from collections import deque
import datetime
from functools import wraps
import inspect
import logging
from uuid import uuid4

from dateutil.parser import parse as parse_date
from lxml import etree

from cubicweb.dataimport.importer import ExtEntity
from cubes.skos import to_unicode
from cubes.saem_ref import eactype_mapping, eacaddress_mapping, eacmaintenancetype_mapping

TYPE_MAPPING = eactype_mapping.copy()
TYPE_MAPPING['human'] = u'person'

ETYPES_ORDER_HINT = ('AgentKind', 'PhoneNumber', 'PostalAddress', 'AuthorityRecord',
                     'AgentPlace', 'Mandate', 'LegalStatus', 'History',
                     'Structure', 'AgentFunction', 'Occupation', 'GeneralContext',
                     'AssociationRelation', 'ChronologicalRelation', 'HierarchicalRelation',
                     'EACResourceRelation', 'ExternalUri', 'EACSource',
                     'Agent', 'Activity')


class InvalidEAC(RuntimeError):
    """EAC input is malformed."""


class InvalidXML(RuntimeError):
    """EAC input has an invalid XML format"""


class MissingTag(RuntimeError):
    """Mandatory tag is missing in EAC input"""

    def __init__(self, tag, tag_parent=None):
        super(MissingTag, self).__init__()
        self.tag = tag
        self.tag_parent = tag_parent


def external_uri(uri):
    values = [unicode(uri)]
    return ExtEntity('ExternalUri', uri, {'uri': set(values), 'cwuri': set(values)})


def filter_none(func):
    """Filter None value from a generator function."""
    def wrapped(*args, **kwargs):
        for x in func(*args, **kwargs):
            if x is not None:
                yield x
    return wraps(func)(wrapped)


def filter_empty(func):
    """Filter out empty ExtEntity (i.e. with empty ``values`` attribute)."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        for extentity in func(self, *args, **kwargs):
            if extentity.values:
                yield extentity
    return wrapper


def elem_maybe_none(func):
    """Method decorator for external entity builder function handling the case
    of `elem` being None.
    """
    if inspect.isgeneratorfunction(func):
        def wrapped(self, elem, *args, **kwargs):
            if elem is None:
                return
            for extentity in func(self, elem, *args, **kwargs):
                yield extentity
    else:
        def wrapped(self, elem, *args, **kwargs):
            if elem is None:
                return None
            else:
                return func(self, elem, *args, **kwargs)
    return wraps(func)(wrapped)


def add_xml_wrap_for(*etypes):
    """Add an `xml_wrap` attribute in ExtEntity's values dictionnary."""
    def decorator(func):
        def wrapped(self, elem):
            objectXMLWrap = self._elem_find(elem, 'eac:objectXMLWrap')
            xmlwrap = None
            if objectXMLWrap is not None:
                nchildren = len(objectXMLWrap)
                if nchildren >= 1:
                    xmlwrap = objectXMLWrap[0]
                if nchildren > 1:
                    msg = self._('multiple children elements found in {0}').format(
                        objectXMLWrap)
                    self.import_log.record_warning(msg, line=objectXMLWrap.line)
            attribute_added = False
            for extentity in func(self, elem):
                if xmlwrap is not None and extentity.etype in etypes:
                    # prevent association of xmlwrap to several extentities.
                    assert not attribute_added, 'xml_wrap attribute already added'
                    extentity.values.setdefault('xml_wrap', set([])).add(
                        etree.tostring(xmlwrap, encoding='utf-8'))
                    attribute_added = True
                yield extentity
        return wraps(func)(wrapped)
    return decorator


def relate_to_record_through(etype, rtype):
    """Add an ``rtype`` relationship from ``etype`` to the imported record."""
    def decorator(func):
        if inspect.isgeneratorfunction(func):
            def wrapper(self, *args, **kwargs):
                for extentity in func(self, *args, **kwargs):
                    if extentity.etype == etype:
                        extentity.values.setdefault(rtype, set()).add(self.record.extid)
                    yield extentity
        else:
            def wrapper(self, *args, **kwargs):
                extentity = func(self, *args, **kwargs)
                if extentity and extentity.etype == etype:
                    extentity.values.setdefault(rtype, set()).add(self.record.extid)
                return extentity
        return wraps(func)(wrapper)
    return decorator


def add_citations_for(etype):
    """Handle import of citation tag for `etype` ExtEntity that is yielded by
    decorated method.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, elem):
            for extentity in func(self, elem):
                if extentity.etype == etype:
                    for citation in self.build_citation(elem):
                        extentity.values.setdefault(
                            'has_citation', set()).add(citation.extid)
                        yield citation
                yield extentity
        return wrapper
    return decorator


def require_tag(tagname):
    """Method decorator handling a mandatory tag within a XML element."""
    def warn(self, elem):
        self.import_log.record_warning(
            self._('expecting a %s tag in element %s, found none') %
            (tagname, elem.tag), line=elem.sourceline)

    def decorator(func):
        # pylint: disable=protected-access
        if inspect.isgeneratorfunction(func):
            def wrapped(self, elem, *args, **kwargs):
                if self._elem_find(elem, tagname) is None:
                    warn(self, elem)
                    return
                for extentity in func(self, elem, *args, **kwargs):
                    yield extentity
        else:
            def wrapped(self, elem, *args, **kwargs):
                if self._elem_find(elem, tagname) is None:
                    warn(self, elem)
                    return None
                return func(self, elem, *args, **kwargs)
        return wraps(func)(wrapped)

    return decorator


def trace_extentity(instance):
    """Decorator for `build_` methods tracing ExtEntities built from a given
    XML element.
    """
    def decorator(func):
        if inspect.isgeneratorfunction(func):
            def wrapper(elem, *args, **kwargs):
                for extentity in func(elem, *args, **kwargs):
                    instance.record_visited(elem, extentity)
                    yield extentity
        else:
            def wrapper(elem, *args, **kwargs):
                extentity = func(elem, *args, **kwargs)
                if extentity is not None:
                    instance.record_visited(elem, extentity)
                return extentity
        return wraps(func)(wrapper)
    return decorator


def equivalent_concept(tagname, etype):
    """Method decorator indicating that a sub-tag may have a vocabularySource attribute indicating
    that a relation to some ExternalUri object should be drown from any entity of type `etype` built
    by decorated method.
    """
    def decorator(func):
        @wraps(func)
        def wrapped(self, elem, *args, **kwargs):
            subelem = self._elem_find(elem, tagname)
            if subelem is not None:
                extid = subelem.attrib.get('vocabularySource')
                if extid is not None:
                    yield external_uri(extid)
            else:
                extid = None

            def update_extentity(extentity):
                if extid is not None and extentity.etype == etype:
                    extentity.values['equivalent_concept'] = set([extid])

            if inspect.isgeneratorfunction(func):
                for extentity in func(self, elem, *args, **kwargs):
                    update_extentity(extentity)
                    yield extentity
            else:
                extentity = func(self, elem, *args, **kwargs)
                update_extentity(extentity)
                yield extentity
        return wrapped
    return decorator


class EACCPFImporter(object):
    """Importer for EAC-CPF data.

    The importer will generate `extid`s using the `extid_generator` function
    if specified or use `uuid.uuid4` to generate unique `extid`s.

    During import the `record` attribute is set to the external entity of the
    imported AuthorityRecord.

    Ref: http://eac.staatsbibliothek-berlin.de/fileadmin/user_upload/schema/cpfTagLibrary.html
    """
    def __init__(self, fpath, import_log, _, extid_generator=None):
        try:
            tree = etree.parse(fpath)
        except etree.XMLSyntaxError as exc:
            raise InvalidXML(str(exc))
        self._ = _
        self._root = tree.getroot()
        self.namespaces = self._root.nsmap.copy()
        self.namespaces['eac'] = self.namespaces.pop(None)
        self.import_log = import_log
        if extid_generator is None:
            def extid_generator():
                return str(uuid4())
        self._gen_extid = extid_generator
        self.record = ExtEntity('AuthorityRecord', None, {})
        # Store a mapping of XML elements to produced ExtEntities
        self._visited = {}

    def __getattribute__(self, name):
        attr = super(EACCPFImporter, self).__getattribute__(name)
        if name.startswith('build_'):
            return trace_extentity(self)(attr)
        return attr

    def record_visited(self, elem, extentity):
        assert extentity.extid, extentity
        self._visited.setdefault(elem, set([])).add(extentity.extid)

    def not_visited(self):
        """Yield (tagname, sourceline) items corresponding to XML elements not
        used to build any ExtEntity.
        """
        visited = self._visited
        ns = self.namespaces['eac']
        queue = deque(self._root)
        # These elements contain other ones which are known to be handled.
        container_tags = ['control', 'cpfDescription', 'identity',
                          'maintenanceHistory', 'sources', 'description',
                          'mandates', 'places', 'occupations', 'relations']
        containers = ['{{{0}}}{1}'.format(ns, tag) for tag in container_tags]
        while queue:
            elem = queue.popleft()
            if (not isinstance(elem, etree._Element) or
                    isinstance(elem, etree._Comment)):
                continue
            if elem in visited:
                continue
            if elem.tag not in containers:
                yield elem.tag.replace('{' + ns + '}', ''), elem.sourceline
            else:
                queue.extend(elem)

    def _elem_find(self, elem, path, method='find'):
        """Wrapper around lxml.etree.Element find* methods to support
        namespaces also for old lxml versions.
        """
        finder = getattr(elem, method)
        try:
            return finder(path, self.namespaces)
        except TypeError:
            # In old lxml, find() does not accept namespaces argument.
            path = path.split(':', 1)
            try:
                ns, path = path
            except ValueError:
                # path has no namespace
                pass
            else:
                path = '{' + self.namespaces[ns] + '}' + path
            return finder(path)

    def _elem_findall(self, *args):
        return self._elem_find(*args, method='findall')

    @filter_empty
    def external_entities(self):
        """Parse a EAC XML file to and yield external entities."""
        # control element.
        control = self._elem_find(self._root, 'eac:control')
        for extentity in self.parse_control(control):
            yield extentity
        # Records (identity tags) are within cpfDescription tag.
        cpf_desc = self._elem_find(self._root, 'eac:cpfDescription')
        if cpf_desc is None:
            raise MissingTag('cpfDescription')
        # identity element.
        identity = self._elem_find(cpf_desc, 'eac:identity')
        if identity is None:
            raise MissingTag('identity', 'cpfDescription')
        for extentity in self.parse_identity(identity):
            yield extentity
        # description element.
        description = self._elem_find(cpf_desc, 'eac:description')
        if description is not None:
            for extentity in self.parse_description(description):
                yield extentity
        # relations element.
        for extentity in self.parse_relations(cpf_desc):
            yield extentity
        # Record is complete.
        self.record_visited(self._root, self.record)
        yield self.record

    def parse_identity(self, identity):
        """Parse the `identity` tag and yield external entities, possibly
        updating record's `values` dict.
        """
        # entityId
        isni = self._elem_find(identity, 'eac:entityId')
        if isni is not None and isni.text:
            self.record_visited(isni, self.record)
            self.record.values['isni'] = set([unicode(isni.text)])
        # entityType
        akind = self._elem_find(identity, 'eac:entityType')
        if akind is None:
            raise MissingTag('entityType', 'identity')
        agent_kind = self.build_agent_kind(akind)
        yield agent_kind
        self.record.values['agent_kind'] = set([agent_kind.extid])
        for name_entry in self._elem_findall(identity, 'eac:nameEntry'):
            self.record_visited(name_entry, self.record)
            if self._elem_find(name_entry, 'eac:authorizedForm') is not None:
                break
        else:
            msg = self._('found no nameEntry tag within an authorizedForm '
                         'element, using %s instead') % name_entry
            self.import_log.record_info(msg, line=name_entry.sourceline)
        part = self._elem_find(name_entry, 'eac:part')
        self.record.values['name'] = set([unicode(part.text)])

    @filter_none
    def parse_description(self, description):
        """Parse the `description` tag and yield external entities, possibly
        updating record's `values` dict.
        """
        # dates.
        daterange = description.xpath('eac:existDates/eac:dateRange',
                                      namespaces=self.namespaces)
        if daterange:
            elem = daterange[0]
            self.record_visited(elem, self.record)
            self.record_visited(elem.getparent(), self.record)
            dates = self.parse_daterange(elem)
            if dates:
                self.record.values.update(dates)
        # address.
        for place in self.find_nested(description, 'eac:place', 'eac:places'):
            for extentity in self.build_place(place):
                yield extentity
        # additional EAC-CPF information.
        for legal_status in self._elem_findall(description, 'eac:legalStatus'):
            for extentity in self.build_legal_status(legal_status):
                yield extentity
        # mandate
        for mandate in self.find_nested(description, 'eac:mandate', 'eac:mandates'):
            for extentity in self.build_mandate(mandate):
                yield extentity
        # history
        for history in self._elem_findall(description, 'eac:biogHist'):
            for extentity in self.build_history(history):
                yield extentity
        # structure
        for structure in self._elem_findall(description, 'eac:structureOrGenealogy'):
            yield self.build_structure(structure)
        # function
        for function in self.find_nested(description, 'eac:function', 'eac:functions'):
            for extentity in self.build_function(function):
                yield extentity
        # occupation
        for occupation in self.find_nested(description, 'eac:occupation',
                                           'eac:occupations'):
            for extentity in self.build_occupation(occupation):
                yield extentity
        # general context
        for context in self._elem_findall(description, 'eac:generalContext'):
            for extentity in self.build_generalcontext(context):
                yield extentity

    def find_nested(self, elem, tagname, innertag):
        """Return a list of element with `tagname` within `element` possibly
        nested within `innertag`.
        """
        all_elems = self._elem_findall(elem, tagname)
        wrapper = self._elem_find(elem, innertag)
        if wrapper is not None:
            all_elems.extend(self._elem_findall(wrapper, tagname))
        return all_elems

    def parse_tag_description(self, elem, tagname='eac:descriptiveNote',
                              attrname='description'):
        """Return a dict with `attrname` and `attrname`_format retrieved from
        a description-like tag.
        """
        elems = self._elem_findall(elem, tagname)
        if len(elems) > 1:
            self.import_log.record_warning(self._(
                'found multiple %s tag within %s element, only one will be '
                'used.') % (tagname, elem.tag), line=elem.sourceline)
        elem = elems[0] if elems else None
        values = {}
        if elem is not None:
            parsed = self.parse_tag_content(elem)
            values.update(zip((attrname, attrname + '_format'),
                              (set([p]) for p in parsed)))
        return values

    def parse_tag_content(self, elem):
        """Parse the content of an element be it text of HTML and return the
        content along with MIME type.
        """
        assert elem is not None, 'unexpected empty element'
        text = elem.text.strip() if elem.text else None
        if text:
            desc, desc_format = unicode(text), u'text/plain'
        else:
            ptag = '{%(eac)s}p' % self.namespaces
            desc = '\n'.join(etree.tostring(child, encoding=unicode,
                                            method='html').strip()
                             for child in elem.iterchildren(ptag)
                             if len(child) != 0 or child.text)
            if desc:
                desc_format = u'text/html'
            else:
                self.import_log.record_warning(self._(
                    'element %s has no text nor children, no content '
                    'extracted') % elem.tag, line=elem.sourceline)
                desc, desc_format = None, None
        return desc, desc_format

    @elem_maybe_none
    def build_agent_kind(self, elem):
        """Build a AgentKind external entity"""
        # Map EAC entity types to our terminolgy.
        kind = TYPE_MAPPING.get(elem.text, elem.text)
        agentkind_id = 'agentkind/' + kind
        return ExtEntity('AgentKind', agentkind_id, {'name': set([unicode(kind)])})

    @elem_maybe_none
    @relate_to_record_through('LegalStatus', 'legal_status_agent')
    @filter_empty
    @add_citations_for('LegalStatus')
    @equivalent_concept('eac:term', 'LegalStatus')
    def build_legal_status(self, elem, **kwargs):
        """Build a `LegalStatus` external entity.

        Extra `kwargs` are passed to `parse_tag_description`.
        """
        values = {}
        term = self._elem_find(elem, 'eac:term')
        if term is not None and term.text:
            values['term'] = set([unicode(term.text)])
        values.update(**self.parse_tag_description(elem, **kwargs))
        dateelem = self._elem_find(elem, 'eac:dateRange')
        dates = self.parse_daterange(dateelem)
        if dates:
            values.update(dates)
        legalstatus = ExtEntity('LegalStatus', self._gen_extid(), values)
        if dateelem is not None:
            self.record_visited(dateelem, legalstatus)
        yield legalstatus

    @elem_maybe_none
    @relate_to_record_through('Mandate', 'mandate_agent')
    @filter_empty
    @add_citations_for('Mandate')
    @equivalent_concept('eac:term', 'Mandate')
    def build_mandate(self, elem, **kwargs):
        """Build a `Mandate` external entity.

        Extra `kwargs` are passed to `parse_tag_description`.
        """
        values = {}
        term = self._elem_find(elem, 'eac:term')
        if term is not None and term.text:
            values['term'] = set([unicode(term.text)])
        values.update(**self.parse_tag_description(elem, **kwargs))
        dateelem = self._elem_find(elem, 'eac:dateRange')
        dates = self.parse_daterange(dateelem)
        if dates:
            values.update(dates)
        yield ExtEntity('Mandate', self._gen_extid(), values)

    @elem_maybe_none
    def build_citation(self, elem):
        """Build a `Citation` external entity."""
        for citation_elem in self._elem_findall(elem, "eac:citation"):
            note = citation_elem.text.strip() if citation_elem.text else u''
            uri = citation_elem.attrib.get('{%(xlink)s}href' % self.namespaces)
            if not note and not uri:
                msg = self._('element {0} has no text nor (valid) link').format(
                    etree.tostring(citation_elem))
                self.import_log.record_warning(msg, line=citation_elem.sourceline)
                return
            values = {}
            if uri:
                values['uri'] = set([unicode(uri)])
            if note:
                values['note'] = set([unicode(note)])
                if u'<span>' in note:
                    values['note_format'] = set([u'text/html'])
            yield ExtEntity('Citation', self._gen_extid(), values)

    @relate_to_record_through('History', 'history_agent')
    @filter_empty
    @add_citations_for('History')
    @elem_maybe_none
    def build_history(self, elem):
        """Build a `History` external entity."""
        desc, desc_format = self.parse_tag_content(elem)
        values = {}
        if desc:
            values.update(text=set([desc]), text_format=set([desc_format]))
        yield ExtEntity('History', self._gen_extid(), values)

    @elem_maybe_none
    @relate_to_record_through('Structure', 'structure_agent')
    def build_structure(self, elem):
        """Build a `Structure` external entity."""
        desc, desc_format = self.parse_tag_content(elem)
        if desc:
            values = {'description': set([desc]),
                      'description_format': set([desc_format])}
            return ExtEntity('Structure', self._gen_extid(), values)

    @relate_to_record_through('AgentPlace', 'place_agent')
    @filter_empty
    @add_citations_for('AgentPlace')
    @equivalent_concept('eac:placeEntry', 'AgentPlace')
    def build_place(self, elem):
        """Build a AgentPlace external entity"""
        values = {}
        role = self._elem_find(elem, 'eac:placeRole')
        if role is not None:
            values['role'] = set([unicode(role.text)])
        entry = self._elem_find(elem, 'eac:placeEntry')
        if entry is not None:
            values['name'] = set([unicode(entry.text)])
        for address in self._elem_findall(elem, 'eac:address'):
            for extentity in self.build_address(address):
                values['place_address'] = set([extentity.extid])
                yield extentity
        yield ExtEntity('AgentPlace', self._gen_extid(), values)

    def build_address(self, elem):
        """Build `PostalAddress`s external entity"""
        address_entity = {}
        for line in self._elem_findall(elem, 'eac:addressLine'):
            if 'localType' in line.attrib:
                attr = dict(eacaddress_mapping).get(line.attrib['localType'])
                if attr:
                    address_entity.setdefault(attr, set()).add(
                        unicode(line.text))
        yield ExtEntity('PostalAddress', self._gen_extid(), address_entity)

    @relate_to_record_through('AgentFunction', 'function_agent')
    @filter_empty
    @add_citations_for('AgentFunction')
    @equivalent_concept('eac:term', 'AgentFunction')
    def build_function(self, elem):
        """Build a `AgentFunction`s external entities"""
        values = {}
        term = self._elem_find(elem, 'eac:term')
        if term is not None:
            values['name'] = set([unicode(term.text)])
        values.update(self.parse_tag_description(elem))
        yield ExtEntity('AgentFunction', self._gen_extid(), values)

    @relate_to_record_through('Occupation', 'occupation_agent')
    @filter_empty
    @add_citations_for('Occupation')
    @equivalent_concept('eac:term', 'Occupation')
    def build_occupation(self, elem):
        """Build a `Occupation`s external entities"""
        values = {}
        term = self._elem_find(elem, 'eac:term')
        if term is not None:
            values['term'] = set([unicode(term.text)])
        values.update(self.parse_tag_description(elem))
        dateelem = self._elem_find(elem, 'eac:dateRange')
        dates = self.parse_daterange(dateelem)
        if dates:
            values.update(dates)
        yield ExtEntity('Occupation', self._gen_extid(), values)

    @relate_to_record_through('GeneralContext', 'general_context_of')
    @filter_empty
    @add_citations_for('GeneralContext')
    def build_generalcontext(self, elem):
        """Build a `GeneralContext` external entity"""
        values = {}
        content, content_format = self.parse_tag_content(elem)
        if content:
            values.update({
                'content': set([content]),
                'content_format': set([content_format]),
            })
        yield ExtEntity('GeneralContext', self._gen_extid(), values)

    @elem_maybe_none
    def parse_daterange(self, elem):
        """Parse a `dateRange` tag and return a dict mapping `start_date` and
        `end_date` to parsed date range.
        """
        values = {}
        for eactag, attrname in zip(('eac:fromDate', 'eac:toDate'),
                                    ('start_date', 'end_date')):
            date = self.parse_date(self._elem_find(elem, eactag))
            if date:
                values[attrname] = set([date])
        return values

    @elem_maybe_none
    def parse_date(self, elem):
        """Parse a date-like element"""
        def record_warning(msg):
            self.import_log.record_warning(msg % {'e': etree.tostring(elem)},
                                           line=elem.sourceline)
        standard_date = elem.attrib.get('standardDate')
        if standard_date:
            date = standard_date
        else:
            for attr in ('notBefore', 'notAfter'):
                if elem.attrib.get(attr):
                    record_warning(self._('found an unsupported %s attribute in date '
                                          'element %%(e)s') % attr)
            # Using element's text.
            date = elem.text
        # Set a default value for month and day; the year should always be
        # given.
        default = datetime.datetime(9999, 1, 1)
        try:
            pdate = parse_date(date, default=default)
        except ValueError as exc:
            record_warning(self._('could not parse date %(e)s'))
            return None
        except Exception as exc:  # pylint: disable=broad-except
            # Usually a bug in dateutil.parser.
            record_warning(self._(
                'unexpected error during parsing of date %%(e)s: %s') %
                to_unicode(exc))
            logger = logging.getLogger('cubicweb.saem_ref')
            logger.exception(self._('unhandled exception while parsing date %r'), date)
            return None
        else:
            if pdate.year == default.year:
                record_warning(self._('could not parse a year from date element %(e)s'))
                return None
            return pdate.date()

    def parse_relations(self, cpf_description):
        """Parse the `relations` tag and yield external entities, possibly
        updating record's `values` dict.
        """
        relations = self._elem_find(cpf_description, 'eac:relations')
        if relations is None:
            return
        # cpfRelation.
        for cpfrel in self._elem_findall(relations, 'eac:cpfRelation'):
            for extentity in self.build_relation(cpfrel):
                yield extentity
        # resourceRelation.
        for rrel in self._elem_findall(relations, 'eac:resourceRelation'):
            for extentity in self.build_resource_relation(rrel):
                yield extentity

    @add_xml_wrap_for('AssociationRelation', 'ChronologicalRelation',
                      'HierarchicalRelation')
    def build_relation(self, elem):
        """Build a relation between records external entity (with proper type)."""
        relationship = elem.attrib.get('cpfRelationType')
        if relationship is None:
            self.import_log.record_warning(self._(
                'found no cpfRelationType attribute in element %s, defaulting '
                'to associative') % etree.tostring(elem),
                line=elem.sourceline)
            relationship = 'associative'
        try:
            # "other_role" (resp. "agent_role") role designates the object of the relation (resp.
            # the agent described in the EAC-CPF instance).
            # See: http://eac.staatsbibliothek-berlin.de/fileadmin/user_upload/schema/cpfTagLibrary.html#cpfRelationType # noqa pylint: disable=line-too-long
            # In case the EAC relation is not qualified, we assume the object is the "parent" (or
            # oldest) in the relation.
            etype, other_role, agent_role = {
                'hierarchical': ('HierarchicalRelation',
                                 'hierarchical_parent', 'hierarchical_child'),
                'hierarchical-parent': ('HierarchicalRelation',
                                        'hierarchical_parent', 'hierarchical_child'),
                'hierarchical-child': ('HierarchicalRelation',
                                       'hierarchical_child', 'hierarchical_parent'),
                'temporal': ('ChronologicalRelation',
                             'chronological_predecessor', 'chronological_successor'),
                'temporal-earlier': ('ChronologicalRelation',
                                     'chronological_predecessor', 'chronological_successor'),
                'temporal-later': ('ChronologicalRelation',
                                   'chronological_successor', 'chronological_predecessor'),
                'associative': ('AssociationRelation',
                                'association_to', 'association_from'),
            }[relationship]
        except KeyError:
            self.import_log.record_warning(self._(
                'unsupported cpfRelationType %s in element %s, skipping')
                % (relationship, etree.tostring(elem)),
                line=elem.sourceline)
            return
        obj_uri = elem.attrib.get('{%(xlink)s}href' % self.namespaces)
        if obj_uri is None:
            self.import_log.record_warning(self._('found a cpfRelation without any object (no '
                                                  'xlink attribute), skipping'),
                                           line=elem.sourceline)
            return
        rentry = self._elem_find(elem, 'eac:relationEntry')
        if rentry is not None and rentry.text.strip():
            yield ExtEntity('AuthorityRecord', obj_uri,
                            {'name': set([unicode(rentry.text)]),
                             'agent_kind': set(['agentkind/unknown-agent-kind'])})
        else:
            yield external_uri(obj_uri)
        values = {agent_role: set([self.record.extid]), other_role: set([obj_uri])}
        dates = self.parse_daterange(
            self._elem_find(elem, 'eac:dateRange'))
        if dates:
            if etype == 'ChronologicalRelation':
                self.import_log.record_warning(
                    self._('ignoring dates on %s since chronological relation '
                           'does not support dates') % etree.tostring(elem)
                )
            else:
                values.update(dates)
        values.update(self.parse_tag_description(elem))
        yield ExtEntity(etype, self._gen_extid(), values)

    @add_xml_wrap_for('EACResourceRelation')
    def build_resource_relation(self, elem):
        """Build a `EACResourceRelation` external entity (along with
        ExternalUri entities).
        """
        obj_uri = elem.attrib.get('{%(xlink)s}href' % self.namespaces)
        if obj_uri is None:
            self.import_log.record_warning(self._(
                'found a resourceRelation without any object (no xlink '
                'attribute), skipping'), line=elem.sourceline)
            return
        yield external_uri(obj_uri)
        values = {
            'resource_relation_resource': set([obj_uri]),
            'resource_relation_agent': set([self.record.extid]),
        }
        resource_role = elem.attrib.get('{%(xlink)s}role' % self.namespaces)
        if resource_role:
            values['resource_role'] = set([unicode(resource_role)])
        agent_role = elem.attrib.get('resourceRelationType')
        if agent_role:
            values['agent_role'] = set([unicode(agent_role)])
        dates = self.parse_daterange(self._elem_find(elem, 'eac:dateRange'))
        if dates:
            values.update(dates)
        values.update(self.parse_tag_description(elem))
        yield ExtEntity('EACResourceRelation', self._gen_extid(), values)

    @filter_none
    def parse_control(self, control):
        """Parse the `control` tag."""
        record_id = self._elem_find(control, 'eac:recordId')
        if record_id is not None and record_id.text and record_id.text.strip():
            self.record.extid = record_id.text.strip()
            self.record_visited(record_id, self.record)
        else:
            self.record.extid = self._gen_extid()
            self.import_log.record_warning(self._(
                'found no recordId element in control tag, using %s as cwuri') %
                self.record.extid)
        for elem in control.xpath('eac:sources/eac:source',
                                  namespaces=self.namespaces):
            for extentity in self.build_source(elem):
                yield extentity
        for elem in control.xpath('eac:maintenanceHistory/eac:maintenanceEvent',
                                  namespaces=self.namespaces):
            for extentity in self.build_maintenance_event(elem):
                yield extentity

    def build_maintenance_event(self, elem):
        """Parse a `maintenanceEvent` tag, yielding a prov:Activity external
        entity along with necessary Records.
        """
        values = {'generated': set([self.record.extid])}
        event_type = self.parse_event_type(self._elem_find(elem, 'eac:eventType'))
        if event_type is not None:
            values['type'] = set([event_type])
        date = self._elem_find(elem, 'eac:eventDateTime')
        if date is not None:
            dtattr = date.attrib.get('standardDateTime')
            if dtattr:
                try:
                    event_date = parse_date(dtattr)
                except ValueError:
                    self.import_log.record_warning(
                        self._('could not parse date from %s') % etree.tostring(date),
                        line=date.sourceline)
                else:
                    values['start'] = set([event_date])
                    values['end'] = set([event_date])
        values.update(self.parse_tag_description(elem, 'eac:eventDescription'))
        agent = self._elem_find(elem, 'eac:agent')
        if agent is not None and agent.text:
            record_id = self._gen_extid()
            yield ExtEntity('Agent', record_id,
                            {'name': set([unicode(agent.text)])})
            values['associated_with'] = set([record_id])
        yield ExtEntity('Activity', self._gen_extid(), values)

    @elem_maybe_none
    def parse_event_type(self, elem):
        """Parse an `eventType` element and try to match a prov:type to build a
        prov:Activity.
        """
        event_type = elem.text.strip() if elem.text else None
        if event_type:
            type_mapping = eacmaintenancetype_mapping.copy()
            type_mapping['derived'] = u'create'
            type_mapping['updated'] = u'modify'
            try:
                event_type = type_mapping[event_type.lower()]
            except KeyError:
                self.import_log.record_warning(self._(
                    'eventType %s does not match the PROV-O vocabulary, respective Activity will '
                    'not have a `type` attribute set.') % event_type, line=elem.sourceline)
                return None
            return event_type

    @relate_to_record_through('EACSource', 'source_agent')
    @filter_empty
    @add_xml_wrap_for('EACSource')
    def build_source(self, elem):
        """Parse a `source` tag, yielding EACSource external entities.
        """
        values = {}
        url = elem.attrib.get('{%(xlink)s}href' % self.namespaces)
        if url is not None:
            values['url'] = set([unicode(url)])
        entry = self._elem_find(elem, 'eac:sourceEntry')
        if entry is not None and entry.text:
            values['title'] = set([unicode(entry.text)])
        values.update(self.parse_tag_description(elem))
        yield ExtEntity('EACSource', self._gen_extid(), values)



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref test for OAI-PMH export"""

from collections import namedtuple
from datetime import datetime, timedelta
from functools import wraps
import time

from dateutil.parser import parse as parse_date
from lxml import etree
import pytz

from logilab.common import tempattr

from cubicweb.devtools.testlib import CubicWebTC

from cubes.oaipmh import isodate
from cubes.oaipmh.entities import OAIComponent
from cubes.oaipmh.views import OAIError, OAIRequest

import testutils


EntityInfo = namedtuple('EntityInfo', ['ark', 'url', 'eid'])


def publish_entity(entity):
    """Fire "publish" transition for `entity`."""
    entity.cw_adapt_to('IWorkflowable').fire_transition('publish')
    entity._cw.commit()


def deprecate_entity(entity):
    """Fire "deprecate" transition for `entity`."""
    entity.cw_adapt_to('IWorkflowable').fire_transition('deprecate')
    entity._cw.commit()


class OAITestMixin(object):

    def oai_component(self, cnx):
        """Return the "oai" component"""
        return self.vreg['components'].select('oai', cnx)


class OAIComponentTC(CubicWebTC, OAITestMixin):

    def test_registered(self):
        with self.admin_access.repo_cnx() as cnx:
            oai = self.oai_component(cnx)
            setspecs = oai.__setspecs__
            self.assertCountEqual(setspecs.keys(),
                                  ['agent', 'conceptscheme', 'profile', 'concept'])
            self.assertCountEqual(setspecs['agent'].keys(), ['role'])
            self.assertCountEqual(setspecs['profile'].keys(),
                                  ['transferring_agent'])
            self.assertEqual(setspecs['conceptscheme'].keys(), [])

    def test_match_setspec(self):
        with self.admin_access.repo_cnx() as cnx:
            oai = self.oai_component(cnx)
            with self.assertRaises(ValueError) as cm:
                oai.match_setspec(u'cwuser')
            self.assertIn('no setspec matching cwuser', str(cm.exception))

    def test_setspecs(self):
        with self.admin_access.repo_cnx() as cnx:
            testutils.organization_unit(
                cnx, u'arch', archival_roles=[u'archival'])
            ark = testutils.organization_unit(
                cnx, u'dep', archival_roles=[u'deposit']).ark
            scheme = testutils.setup_scheme(cnx, u'test', u'lab')
            cnx.commit()
            publish_entity(scheme)
            expected = (
                'agent',
                'agent:role:producer', 'agent:role:deposit',
                'agent:role:archival', 'agent:role:control',
                'agent:role:enquirer', 'agent:role:seda-actor',
                'conceptscheme',
                'profile', 'profile:transferring_agent:%s' % ark,
                'concept', 'concept:in_scheme:%s' % scheme.ark
            )
            oai = self.oai_component(cnx)
            setspecs = [x[0] for x in oai.setspecs()]
            self.assertCountEqual(setspecs, expected)


def setup_agents(cnx):
    """Create three agents (alice, bobby and draft), publish alice and bobby but not draft, and return
    the (ark, url, eid) attributes of alice and bobby.
    """
    alice = testutils.organization_unit(cnx, u'alice',
                                        archival_roles=[u'producer'])
    cnx.commit()
    publish_entity(alice)
    bobby = testutils.organization_unit(cnx, u'bobby',
                                        archival_roles=[u'enquirer'])
    cnx.commit()
    publish_entity(bobby)
    testutils.organization_unit(cnx, u'draft', archival_roles=[u'enquirer'])
    cnx.commit()
    return (
        EntityInfo(alice.ark, alice.absolute_url(), alice.eid),
        EntityInfo(bobby.ark, bobby.absolute_url(), bobby.eid),
    )


class OAIRequestTC(CubicWebTC):

    def test_rset_from_setspec(self):
        """Test `rset_from_setspec` function for exceptions."""
        with self.admin_access.repo_cnx() as cnx:
            return setup_agents(cnx)
        with self.admin_access.web_request() as req:
            for setspec, msg in [
                ('agent:zen', 'invalid set specifier agent:zen'),
                ('agent:role:control', (
                    'The combination of the values of the from, until, and '
                    'set arguments results in an empty list.')),
                ('conceptscheme:haha', 'invalid set specifier conceptscheme:haha'),
                ('scregneugneu', 'invalid set specifier scregneugneu'),
            ]:
                with self.subTest(setspec=setspec):
                    oairq = OAIRequest('http://testing.fr', setspec=setspec)
                    with self.assertRaises(OAIError) as cm:
                        oairq.rset_from_setspec(req)
                    self.assertEqual(cm.exception.errors, {'noRecordsMatch': msg})

    def test_rset_from_setspec_token(self):
        with self.admin_access.repo_cnx() as cnx:
            alice, bobby = setup_agents(cnx)

        def oairq(**kwargs):
            return OAIRequest('http://testing.fr', setspec='agent', **kwargs)

        with self.admin_access.web_request() as req:
            with tempattr(OAIComponent, 'max_result_size', 1):
                rset, token = oairq().rset_from_setspec(req)
                self.assertEqual(len(rset), 1)
                self.assertEqual(token, str(bobby.eid))
                rset, token = oairq(resumption_token=str(bobby.eid)).rset_from_setspec(req)
                self.assertEqual(len(rset), 1)
                self.assertIsNone(token)
            # All results are returned, the resumptionToken is None.
            rset, token = oairq().rset_from_setspec(req)
            self.assertEqual(len(rset), 2)
            self.assertIsNone(token)
            # and this, even if the user specified an input token.
            rset, token = oairq(resumption_token='1').rset_from_setspec(req)
            self.assertEqual(len(rset), 2)
            self.assertIsNone(token)


def no_validate_xml(method):
    """Disable XML schema validation, often because the underlying metadata
    part of the response (RDF, XSD) is not validable (or we don't know how to
    do it).
    """
    @wraps(method)
    def wrapper(self):
        self._validate_xml = False
        self._debug_xml = True
        return method(self)
    return wrapper


class OAIPMHViewsTC(CubicWebTC, OAITestMixin, testutils.XmlTestMixin):

    _validate_xml = True
    _debug_xml = False

    def oai_request(self, req, **formparams):
        req.form.update(formparams)
        out = self.app_handle_request(req, 'oai')
        if self._validate_xml:
            self.assertXmlValid(out, self.datapath('OAI-PMH.xsd'), debug=self._debug_xml)
        return out

    def _setup_agents(self, **kwargs):
        with self.admin_access.repo_cnx() as cnx:
            return setup_agents(cnx, **kwargs)

    def test_xml_attributes_and_namespaces(self):
        """Check XML attributes and namespace declaration of the response."""
        with self.admin_access.web_request() as req:
            # simple request, generating an error but enough to get a properly
            # formatter response.
            result = self.oai_request(req)
            xml = etree.fromstring(result)
            nsmap = {None: 'http://www.openarchives.org/OAI/2.0/',
                     'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
            self.assertEqual(xml.nsmap, nsmap)
            attrib = {
                '{http://www.w3.org/2001/XMLSchema-instance}schemaLocation': ' '.join(
                    ['http://www.openarchives.org/OAI/2.0/',
                     'http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd']),
            }
            self.assertEqual(xml.attrib, attrib)

    def test_responsedate(self):
        """Check <responseDate>"""
        with self.admin_access.web_request() as req:
            # Simple request, generating an error but enough to get a properly
            # formatter response.
            now = datetime.utcnow()
            result = self.oai_request(req)
        # Check tzinfo abbreviation is present (Z for UTC).
        self.assertIn('<responseDate>{0}</responseDate>'.format(isodate(now)),
                      result)

    def _setup_agents_deleted(self):
        alice, bobby = self._setup_agents()
        with self.admin_access.repo_cnx() as cnx:
            entity = cnx.entity_from_eid(bobby.eid)
            # ensure transition date is greater by at least 1s
            time.sleep(1)
            date_before = entity.modification_date.replace(tzinfo=pytz.utc)
            deprecate_entity(entity)
        return alice, bobby, date_before

    def _setup_schemes(self):
        with self.admin_access.repo_cnx() as cnx:
            pscheme = testutils.setup_scheme(cnx, u'public',
                                             u'public label', u'other public label')
            # draft ConceptScheme
            dscheme = testutils.setup_scheme(cnx, u'draft',
                                             u'draft label', u'other draft label')
            cnx.commit()
            publish_entity(pscheme)
            return (
                EntityInfo(pscheme.ark, pscheme.absolute_url(), pscheme.eid),
                EntityInfo(dscheme.ark, dscheme.absolute_url(), dscheme.eid),
            )

    def test_noverb(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req)
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn('<error code="badVerb">no verb specified</error',
                      result)

    def test_badverb(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='coucou')
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn('<error code="badVerb">illegal verb "coucou"</error>',
                      result)

    # some of our setspecs (those with an ARK inside are not valid)
    @no_validate_xml
    def test_listsets(self):
        pscheme, dscheme = self._setup_schemes()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListSets')
        self.assertIn('<request verb="ListSets">http://testing.fr/cubicweb/oai</request>',
                      result)
        for spec in ('agent', 'conceptscheme', 'profile', 'concept'):
            self.assertIn('<setSpec>{0}</setSpec>'.format(spec),
                          result)
        self.assertIn(('<setSpec>agent:role:deposit</setSpec>'
                       '<setName>An organization unit with deposit archival role</setName>'),
                      result)
        self.assertIn('<setSpec>concept:in_scheme:{0}</setSpec>'.format(pscheme.ark),
                      result)
        self.assertNotIn(dscheme.ark, result)

    def test_listidentifiers_noset(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers')
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn(('<error code="badArgument">'
                       'ListIdentifiers verb requires "set" restriction'
                       '</error>'),
                      result)

    def _setup_profile(self, **kwargs):
        """Create a "published" SEDAProfile and return its identifier"""
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx, **kwargs)
            cnx.commit()
            publish_entity(profile)
            return profile.ark

    def test_resumptiontoken_listidentifiers(self):
        alice, bobby = self._setup_agents()

        def check_request(expected_arks, token=None):
            with self.admin_access.web_request() as req:
                result = self.oai_request(
                    req, verb='ListIdentifiers', set='agent',
                    resumptionToken=token)
                # Ensure there are as many <identifier> tag than expected items.
                self.assertEqual(result.count('<identifier>'), len(expected_arks))
                for ark in expected_arks:
                    self.assertIn(
                        '<identifier>ark:/{0}</identifier>'.format(ark), result)
                return result

        # less results than pagination limit
        result = check_request([alice.ark, bobby.ark])
        # All results returned, no resumptionToken is request -> no
        # resumptionToken in response.
        self.assertNotIn('resumptionToken', result)
        with tempattr(OAIComponent, 'max_result_size', 1):
            # max_result_size = 1, 1st batch
            result = check_request([alice.ark])
            # token + expirationDate
            root = etree.fromstring(result)
            token = root.find('.//{%s}resumptionToken' % root.nsmap[None])
            date = parse_date(token.attrib['expirationDate'])
            self.assertLess(datetime.now(pytz.utc) + timedelta(hours=1) - date,
                            timedelta(seconds=1))
            self.assertEqual(token.text, str(bobby.eid))
            # max_result_size = 1, 2nd batch
            result = check_request([bobby.ark], token=str(bobby.eid))
            # max_result_size = 1, no more result
            # No more results -> empty resumptionToken.
            self.assertIn('<resumptionToken/>', result)
            # Empty result, probably a badResumptionToken.
            result = check_request([], token=str(bobby.eid + 1))
            self.assertIn('badResumptionToken', result)

    def test_listidentifiers(self):
        alice, bobby = [x.ark for x in self._setup_agents()]
        profile_id = self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent')
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(bobby), result)
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(alice), result)
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            self.assertNotIn(draft, result)
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:role:producer')
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(alice),
                          result)
            self.assertNotIn('<identifier>ark:/{0}</identifier>'.format(bobby),
                             result)
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:role:producer')
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(alice), result)
            self.assertNotIn('<identifier>ark:/{0}</identifier>'.format(bobby), result)
            result = self.oai_request(req, verb='ListIdentifiers', set='conceptscheme')
            result = self.oai_request(req, verb='ListIdentifiers', set='profile')
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(profile_id), result)

    def test_listidentifiers_setspec_wfstate(self):
        """Check workflow state condition is statisfied with a multi-level
        setspec `agent:role:enquirer`.
        """
        alice, bobby = [x.ark for x in self._setup_agents()]
        self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers',
                                      set='agent:role:enquirer')
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(bobby), result)
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            self.assertNotIn(draft, result)

    def test_listidentifiers_deleted(self):
        """Check <header> element (status attribute, datestamp) for deleted
        records in ListIdentifiers response.
        """
        alice, bobby, date_before = self._setup_agents_deleted()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='agent')
            self.assertIn(
                '<identifier>ark:/{0}</identifier>'.format(bobby.ark), result)
            self.assertIn('<header status="deleted">', result)
            root = etree.fromstring(result)
            ns = root.nsmap[None]
            # only one <header status="..."> element
            header, = root.findall('.//{%s}header[@status]' % ns)
            datestamp = header.find('{%s}datestamp' % ns).text
            date = parse_date(datestamp)
            self.assertGreater(date, date_before.replace(microsecond=0))

    def test_listrecords_noset(self):
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords')
        self.assertIn('<request>http://testing.fr/cubicweb/oai</request>',
                      result)
        self.assertIn(('<error code="badArgument">'
                       'ListRecords verb requires "set" restriction'
                       '</error>'),
                      result)

    @no_validate_xml
    def test_listrecords(self):
        alice, bobby = self._setup_agents()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords', set='agent')
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(bobby.url), result)
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            self.assertNotIn(draft, result)
            result = self.oai_request(
                req, verb='ListRecords', set='agent:role:producer')
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)
            self.assertNotIn(bobby.url, result)

    @no_validate_xml
    def test_listrecord_deleted(self):
        """Check absence of <metadata> or <about> elements for deleted
        records in ListRecords response.
        """
        alice, bobby, _ = self._setup_agents_deleted()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords', set='agent')
            # no metadata for bobby
            self.assertNotIn(
                '<rdf:Description rdf:about="{0}">'.format(bobby.url), result)
            # but record present
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(bobby.ark), result)
            self.assertIn(
                '<rdf:Description rdf:about="{0}">'.format(alice.url), result)

    @no_validate_xml
    def test_listrecords_profile(self):
        self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListRecords', set='profile')
        # Just ensure a SEDA profile (xsd) is present in response.
        self.assertIn('<metadata><xsd:schema', result)
        self.assertIn('fr:gouv:ae:archive:draft:standard_echange_v0.2', result)

    @no_validate_xml
    def test_listidentifiers_profile_setspec(self):
        """Check setspec restriction for SEDAProfile."""
        profile_id = self._setup_profile()
        with self.admin_access.cnx() as cnx:
            tagent_eid = testutils.organization_unit(
                cnx, u'blah', archival_roles=[u'deposit']).eid
            cnx.commit()
        profile2_id = self._setup_profile(seda_transferring_agent=tagent_eid)
        with self.admin_access.web_request() as req:
            transferring_agent_ark = req.execute(
                'Any K WHERE P ark "%s", P seda_transferring_agent A, A ark K'
                % profile_id)[0][0]
            result = self.oai_request(
                req, verb='ListIdentifiers',
                set='profile:transferring_agent:' + transferring_agent_ark)
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(profile_id), result)
            self.assertNotIn(profile2_id, result)
        # Check that without a setspec, we get all profiles.
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='ListIdentifiers', set='profile')
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(profile_id), result)
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(profile2_id), result)

    @no_validate_xml
    def test_getrecord_profile(self):
        profile_id = self._setup_profile()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='GetRecord')
            self.assertIn('<error code="badArgument">', result)
            result = self.oai_request(req, verb='GetRecord', identifier='999')
            self.assertIn('<error code="idDoesNotExist"', result)
            result = self.oai_request(req, verb='GetRecord', identifier='ark:/' + profile_id)
            self.assertIn('<metadata><xsd:schema', result)
            self.assertIn('fr:gouv:ae:archive:draft:standard_echange_v0.2', result)

    @no_validate_xml
    def test_getrecord_organizationunit(self):
        self._debug_xml = True
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.organization_unit(cnx, u'toto')
            cnx.commit()
            publish_entity(agent)
            ark = agent.ark
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='GetRecord', identifier='ark:/' + ark)
            self.assertIn('name>toto</', result)
            self.assertIn('<rdf:type rdf:resource="http://xmlns.com/foaf/0.1/Organization"/>',
                          result)

    def test_getrecord_deleted(self):
        """Check absence of <metadata> or <about> elements for deleted
        records in GetRecord response.
        """
        alice, bobby, _ = self._setup_agents_deleted()
        with self.admin_access.web_request() as req:
            result = self.oai_request(req, verb='GetRecord', identifier='ark:/' + bobby.ark)
            # record is present
            self.assertIn('<identifier>ark:/{0}</identifier>'.format(bobby.ark), result)
            # but not data
            self.assertNotIn('metadata', result)
            self.assertNotIn('rdf', result)

    def test_getrecord_draft(self):
        self._setup_agents()
        with self.admin_access.web_request() as req:
            draft = req.execute('Any A WHERE X ark A, X name "draft"')[0][0]
            result = self.oai_request(req, verb='GetRecord', identifier='ark:/' + draft)
            self.assertIn('idDoesNotExist', result)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref custom behaviour for search"""

from logilab.mtconverter import TransformError, xml_escape

from cubicweb.uilib import cut
from cubicweb.utils import json_dumps
from cubicweb.predicates import is_instance
from cubicweb.view import EntityView
from cubicweb.web import facet
from cubicweb.web.views import magicsearch, baseviews, basecontrollers, ajaxcontroller

_ = unicode


# highlighting #################################################################

def highlight_keywords(cw):
    """Highlight the form keywords (if any) on page loading"""
    if 'highlight' in cw.form:
        # as we are handling this upstream, directly from the controller and
        # not from a view, jquery hasn't been loaded yet.
        # As jquery.highlight relies on it, we have to add it first.
        cw.add_js((u'jquery.js', u'jquery.highlight.js'))
        cw.add_onload(
            u'$(document).ready(function(){$("h1, h2, h3, h4, h5, '
            'table tbody td").highlight(%s);});' % json_dumps(cw.form['highlight']))


class FullTextTranslator(magicsearch.FullTextTranslator):
    """Full text translator adding a 'highlight' element to _cw.form"""

    def preprocess_query(self, uquery):
        self._cw.form['highlight'] = uquery
        return super(FullTextTranslator, self).preprocess_query(uquery)


class TextSearchResultView(baseviews.TextSearchResultView):
    """Overloaded so that it adds a 'highlight' keyword to the entities URL"""

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.complete_entity(row, col)
        self.w(entity.view('searchresult'))
        searched = self.cw_rset.searched_text()
        if searched is None:
            return
        searched = searched.lower()
        highlighted = '<b>%s</b>' % searched
        for attr in entity.e_schema.indexable_attributes():
            try:
                value = xml_escape(entity.printable_value(attr, format='text/plain').lower())
            except TransformError:
                continue
            if searched in value:
                contexts = []
                for ctx in value.split(searched):
                    if len(ctx) > 30:
                        contexts.append(u'...' + ctx[-30:])
                    else:
                        contexts.append(ctx)
                value = u'\n' + highlighted.join(contexts)
                self.w(value.replace('\n', '<br/>'))


class EntitySearchResultView(EntityView):
    """Overloaded so that it adds an 'highlight' keyword to the URL"""

    __regid__ = 'searchresult'

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        desc = cut(entity.dc_description(), 50)
        if 'highlight' in self._cw.form:
            url = entity.absolute_url(highlight=self._cw.form['highlight'])
        else:
            url = entity.absolute_url()
        self.w(u'<a href="%s" title="%s">%s</a>' % (
            xml_escape(url), xml_escape(desc), xml_escape(entity.dc_title())))


# Beware: this code overrides the monkeypatch of the bootstrap cube.
# See tickets https://www.cubicweb.org/ticket/5098473 and
# https://www.cubicweb.org/ticket/5102924
class SameETypeListView(baseviews.SameETypeListView):
    """Overloaded so that it takes into account subvid form param"""

    def call(self, **kwargs):
        """display a list of entities by calling their <item_vid> view"""
        subvid = self._cw.form.pop('subvid') if 'subvid' in self._cw.form else None
        if subvid is not None:
            self.item_vid = subvid
        super(SameETypeListView, self).call(**kwargs)


class ViewController(basecontrollers.ViewController):
    """Overloaded view controller dealing with 'highlight' keyword in the URL

    Highlight in the generated view the words passed to the url with 'highlight' keyword
    """

    def _select_view_and_rset(self, rset):
        highlight_keywords(self._cw)
        return super(ViewController, self)._select_view_and_rset(rset)


class AjaxController(ajaxcontroller.AjaxController):

    def publish(self, rset=None):
        highlight_keywords(self._cw)
        return super(AjaxController, self).publish(rset)


# facets #######################################################################

class AgentStartDateFacet(facet.DateRangeFacet):
    __regid__ = 'saem_ref.agent_start_date_facet'
    __select__ = facet.DateRangeFacet.__select__ & is_instance('Agent')
    rtype = 'start_date'


class AgentEndDateFacet(facet.DateRangeFacet):
    __regid__ = 'saem_ref.agent_end_date_facet'
    __select__ = facet.DateRangeFacet.__select__ & is_instance('Agent')
    rtype = 'end_date'


class AgentKindFacet(facet.RelationAttributeFacet):
    __regid__ = 'saem_ref.agent_kind_facet'
    __select__ = facet.RelationAttributeFacet.__select__ & is_instance('Agent')
    rtype = 'agent_kind'
    role = 'subject'
    target_attr = 'name'


class AgentRoleFacet(facet.RelationAttributeFacet):
    __regid__ = 'saem_ref.agent_role_facet'
    __select__ = facet.RelationAttributeFacet.__select__ & is_instance('Agent')
    rtype = 'archival_role'
    role = 'subject'
    target_attr = 'name'


################################################################################

def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__,
                      (FullTextTranslator, TextSearchResultView,
                       SameETypeListView, ViewController, AjaxController))
    vreg.register_and_replace(FullTextTranslator, magicsearch.FullTextTranslator)
    vreg.register_and_replace(TextSearchResultView, baseviews.TextSearchResultView)
    vreg.register_and_replace(SameETypeListView, baseviews.SameETypeListView)
    vreg.register_and_replace(ViewController, basecontrollers.ViewController)
    vreg.register_and_replace(AjaxController, ajaxcontroller.AjaxController)



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref site customizations"""

import pytz

from logilab.common.date import ustrftime
from logilab.common.decorators import monkeypatch

from cubicweb import cwvreg
from cubicweb.cwconfig import register_persistent_options
from cubicweb.uilib import PRINTERS

from cubes.skos import rdfio
from cubes.skos.ccplugin import ImportSkosData

from cubes.saem_ref import cwuri_url
# this import is needed to take account of pg_trgm monkeypatches
# while executing cubicweb-ctl commands (db-rebuild-fti)
from cubes.saem_ref import pg_trgm  # noqa pylint: disable=unused-import

_ = unicode


@monkeypatch(rdfio.RDFGraphGenerator, methodname='equivalent_concept_uris')
@staticmethod
def equivalent_concept_uris(entity):
    yield cwuri_url(entity)


# deactivate date-format and datetime-format cw properties. This is because we do some advanced date
# manipulation such as allowing partial date and this is not generic enough to allow arbitrary
# setting of date and time formats

base_user_property_keys = cwvreg.CWRegistryStore.user_property_keys


@monkeypatch(cwvreg.CWRegistryStore)
def user_property_keys(self, withsitewide=False):
    props = base_user_property_keys(self, withsitewide)
    return [prop for prop in props if prop not in ('ui.date-format', 'ui.datetime-format')]

# customize display of TZDatetime

register_persistent_options((
    ('timezone',
     {'type': 'choice',
      'choices': pytz.common_timezones,
      'default': 'Europe/Paris',
      'help': _('timezone in which time should be displayed'),
      'group': 'ui', 'sitewide': True,
      }),
))


def print_tzdatetime_local(value, req, *args, **kwargs):
    tz = pytz.timezone(req.property_value('ui.timezone'))
    value = value.replace(tzinfo=pytz.utc).astimezone(tz)
    return ustrftime(value, req.property_value('ui.datetime-format'))

PRINTERS['TZDatetime'] = print_tzdatetime_local


# configure c-c skos-import command's factories to use with proper metadata generator ##############

def _massive_store_factory(cnx):
    from cubicweb.dataimport.massive_store import MassiveObjectStore
    from cubes.saem_ref.sobjects.skos import SAEMMetadataGenerator
    return MassiveObjectStore(cnx, metagen=SAEMMetadataGenerator(cnx))


def _nohook_store_factory(cnx):
    from cubicweb.dataimport.stores import NoHooRQLObjectStore
    from cubes.saem_ref.sobjects.skos import SAEMMetadataGenerator
    return NoHooRQLObjectStore(cnx, metagen=SAEMMetadataGenerator(cnx))


ImportSkosData.cw_store_factories['massive'] = _massive_store_factory
ImportSkosData.cw_store_factories['nohook'] = _nohook_store_factory


####################################################################################################
# temporary monkey-patches #########################################################################
####################################################################################################

from yams.constraints import Attribute, BoundaryConstraint, cstr_json_loads  # noqa


@monkeypatch(BoundaryConstraint, methodname='deserialize')
@classmethod
def deserialize(cls, value):
    """simple text deserialization"""
    try:
        values = cstr_json_loads(value)
        return cls(**values)
    except ValueError:
        try:
            value, msg = value.split('\n', 1)
        except ValueError:
            msg = None
        op, boundary = value.split(' ', 1)
        return cls(op, eval(boundary), msg or None)



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref application package

Référenciel de Système d'Archivage Électronique Mutualisé
"""

from functools import partial

from logilab.common.registry import objectify_predicate

from cubicweb import neg_role
from cubes.compound import CompositeGraph


# Composite graph definitions used in permissions setup.
PERMISSIONS_GRAPHS = dict.fromkeys(
    ('AuthorityRecord', 'ConceptScheme', 'ProfileDocument', 'ProfileArchiveObject'),
    partial(CompositeGraph, skiprtypes=('generated', 'used')))

OUTER_RELATIVE_PERM_RELATIONS = set([
    'seda_file_type_code_value', 'seda_character_set_code_value',
    'seda_document_type_code_value', 'seda_appraisal_rule_code_value',
    'seda_access_restriction_code_value', 'seda_description_level_value',
    'seda_keyword_scheme', 'seda_keyword_value',
    # those are not outer but non-mandatory...
    'seda_parent', 'seda_oldest_date', 'seda_latest_date',
])


# EAC mappings

eactype_mapping = {'corporateBody': u'authority',
                   'person': u'person',
                   'family': u'family'}
eacmaintenancetype_mapping = {"created": u'create',
                              "revised": u'modify'}
# Order matters for this one in order to export correctly
eacaddress_mapping = [('StreetName', 'street'), ('PostCode', 'postalcode'), ('CityName', 'city')]


def optional_relations(schema, graph_structure):
    """Return a dict with optional relations information in a CompositeGraph.

    Keys are names of entity types in the graph for which a relation type has
    no mandatory (cardinality in '1+') relation definitions and values is a
    set of respective `(rtype, role)` tuples.
    """
    optionals = dict()
    for etype, relations in graph_structure.iteritems():
        for (rtype, role), targets in relations.iteritems():
            for target in targets:
                rdef = schema[rtype].role_rdef(etype, target, role)
                if rdef.role_cardinality(role) in '1+':
                    break
            else:
                optionals.setdefault(etype, set()).add((rtype, role))
    return optionals


def mandatory_rdefs(schema, graph_structure):
    """Yield non-optional relation definitions (and the role of the parent in
    the relation) in a graph structure.
    """
    visited = set()
    for etype, relations in graph_structure.iteritems():
        for (rtype, role), targets in relations.iteritems():
            for target in targets:
                rdef = schema[rtype].role_rdef(etype, target, role)
                if rdef in visited:
                    continue
                visited.add(rdef)
                if rdef.role_cardinality(role) in '1+':
                    yield rdef, neg_role(role)


def cwuri_url(entity):
    """Return an absolute URL for entity's cwuri, handling case where ark is directly used, and so
    URL should be generated from it.
    """
    cwuri = entity.cwuri
    if cwuri.startswith('ark:'):
        cwuri = entity._cw.build_url(cwuri)
    return cwuri


@objectify_predicate
def user_has_authority(cls, req, **kwargs):
    """Return 1 if the user is associated to an authority."""
    return len(req.user.authority)


@objectify_predicate
def user_has_naa(cls, req, **kwargs):
    """Return 1 if the user is associated to an authority with a NAA configured."""
    return 1 if req.user.naa is not None else 0



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import CubicWebTC


class ArkGeneratorTC(CubicWebTC):

    def test_eid(self):
        with self.admin_access.repo_cnx() as cnx:
            for etype, expected in (
                ("AuthorityRecord", u"r000000042"),
                ("SEDAProfile", u"p000000042"),
                ("Concept", u"c000000042"),
                ("ConceptScheme", u"v000000042"),
                ("Organization", u"o000000042"),
                ("OrganizationUnit", u"ou000000042"),
                ("Agent", u"oa000000042"),
                ("CWUser", u"000000042"),
            ):
                generator = self.vreg['adapters'].select('IARKGenerator', cnx, eid=42, etype=etype,
                                                         naa_what='saemref-test')
                self.assertEqual(generator.generate_ark(), 'saemref-test/{0}'.format(expected))

    def test_ext_identifier(self):
        with self.admin_access.repo_cnx() as cnx:
            generator = self.vreg['adapters'].select('IARKGenerator', cnx,
                                                     naa_what='saemref-test')
            self.assertEqual(generator.generate_ark(), 'saemref-test/a000000001')


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""SKOS import overrides"""

import sys

from cubicweb.dataimport.stores import MetadataGenerator, NoHookRQLObjectStore

from cubes.skos import sobjects as skos

from cubes.saem_ref.hooks import set_ark_and_cwuri


def _store(cnx, **kwargs):
    metagen = SAEMMetadataGenerator(cnx, **kwargs)
    # XXX detect if import is started through source-sync, because the massive store isn't usable
    # when other connections are active on the database
    if 'source-sync' in sys.argv and cnx.repo.system_source.dbdriver == 'postgres':
        from cubicweb.dataimport.massive_store import MassiveObjectStore
        return MassiveObjectStore(cnx, metagen=metagen, eids_seq_range=1000)
    else:
        return NoHookRQLObjectStore(cnx, metagen=metagen)


class SAEMMetadataGenerator(MetadataGenerator):
    """SAEM specific metadata generator, necessary when no hooks are activated (e.g. skos import) to
    handle ARK generation and setting of the initial state.
    """

    def __init__(self, cnx, baseurl=None, source=None, naa_what=None):
        super(SAEMMetadataGenerator, self).__init__(cnx, baseurl, source)
        if naa_what is None and source is not None:
            source_entity = cnx.find('CWSource', eid=source.eid).one()
            if source_entity.reverse_through_cw_source:
                skos_source = source_entity.reverse_through_cw_source[0]
                if skos_source.ark_naa:
                    naa_what = skos_source.ark_naa[0].what
        self.naa_what = naa_what

    def etype_rels(self, etype):
        etype_rels = super(SAEMMetadataGenerator, self).etype_rels(etype)
        if etype == 'ConceptScheme':
            etype_rels.append('in_state')
        return etype_rels

    def gen_in_state(self, etype):
        wf_state = self._cnx.execute('Any S WHERE ET default_workflow WF, ET name %(etype)s, '
                                     'WF initial_state S', {'etype': etype}).one()
        return wf_state.eid

    def entity_attrs(self, etype):
        entity_attrs = super(SAEMMetadataGenerator, self).entity_attrs(etype)
        if etype in ('ConceptScheme', 'Concept'):
            entity_attrs.insert(0, 'ark')  # insert before cwuri
        return entity_attrs

    def gen_ark(self, etype, eid, attrs):
        """ARK generation callback"""
        set_ark_and_cwuri(self._cnx, eid, etype, attrs, naa_what=self.naa_what)
        # returning value, even if set by the function, is necessary due to the meta generator
        # implementation
        return attrs['ark']


class RDFSKOSImportService(skos.RDFSKOSImportService):
    """service overriden from skos cube to provide functionnalities handled in saem_ref's hooks"""

    def _do_import(self, stream, import_log, **kwargs):
        return super(RDFSKOSImportService, self)._do_import(stream, import_log,
                                                            store=_store(self._cw), **kwargs)


class LCSVSKOSImportService(skos.LCSVSKOSImportService):
    """service overriden from skos cube to provide functionnalities handled in saem_ref's hooks"""

    def _do_import(self, stream, import_log, **kwargs):
        scheme = self._cw.find('ConceptScheme', cwuri=kwargs['scheme_uri']).one()
        naa_what = scheme.cw_adapt_to('IArkNAALocator').naa_what()
        store = _store(self._cw, naa_what=naa_what)
        return super(LCSVSKOSImportService, self)._do_import(stream, import_log,
                                                             store=store, **kwargs)


class SKOSParser(skos.SKOSParser):
    """parser overriden from skos cube to provide functionnalities handled in saem_ref's hooks"""

    def _do_import(self, url, raise_on_error):
        return super(SKOSParser, self)._do_import(url, raise_on_error=raise_on_error,
                                                  store=_store(self._cw, source=self.source))


def registration_callback(vreg):
    vreg.register_and_replace(RDFSKOSImportService, skos.RDFSKOSImportService)
    vreg.register_and_replace(LCSVSKOSImportService, skos.LCSVSKOSImportService)
    vreg.register_and_replace(SKOSParser, skos.SKOSParser)



import os
import os.path as osp
from cubicweb.dataimport.importer import SimpleImportLog


print('importing some eac files')
EAC_DIR = osp.join(osp.dirname(__file__), os.pardir, 'test', 'data', 'EAC')
for fname in ('FRAD033_EAC_00001.xml', 'FRAD033_EAC_00003.xml', 'FRAD033_EAC_00071.xml'):
    print fname
    import_log = SimpleImportLog(fname)
    cnx.call_service('saem_ref.eac-import', stream=open(osp.join(EAC_DIR, fname)),
                     import_log=import_log, raise_on_error=True)
    cnx.commit()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for hooks"""

from datetime import datetime, timedelta
from time import mktime
import unittest

import pytz

from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref.hooks import extract_ark

import testutils

YESTERDAY = datetime.now(tz=pytz.utc) - timedelta(days=1)


class SAEMRefHooksTC(CubicWebTC):

    def assertMDNow(self, entity):
        entity.cw_clear_all_caches()
        self.assertAlmostEqual(mktime(entity.modification_date.timetuple()),
                               mktime(datetime.now(tz=pytz.utc).timetuple()), delta=60)

    def resetMD(self, cnx, *entities):
        for entity in entities:
            with cnx.deny_all_hooks_but():
                entity.cw_set(modification_date=YESTERDAY)
        cnx.commit()

    def test_reset_md(self):
        """Ensure `resetMD` method above works."""
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'bob')
            cnx.commit()
            self.resetMD(cnx, agent)
            agent.cw_clear_all_caches()
            self.assertEqual(agent.modification_date, YESTERDAY)

    def test_sync_scheme_md(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus',
                                       ark_naa=testutils.naa(cnx))
            cnx.commit()
            self.resetMD(cnx, scheme)
            c1 = scheme.add_concept(u'concept 1', language_code=u'fr')
            cnx.commit()
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme)
            lab = cnx.create_entity('Label', label=u'concept 1.1', language_code=u'en',
                                    label_of=c1)
            cnx.commit()
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme)
            lab.cw_set(label=u'concept 1.1.1')
            cnx.commit()
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme, c1)
            c2 = c1.add_concept(u'sub concept')
            cnx.commit()
            self.assertMDNow(c1)
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme, c1, c2)
            c2.add_concept(u'sub-sub concept')
            cnx.commit()
            self.assertMDNow(c2)
            self.assertMDNow(c1)
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme, c1, c2)
            c2.preferred_label[0].cw_set(label=u'sub concept 2')
            cnx.commit()
            self.assertMDNow(c2)
            self.assertMDNow(c1)
            self.assertMDNow(scheme)

    def test_sync_profile_md(self):
        with self.admin_access.repo_cnx() as cnx:
            testutils.seda_scheme(cnx, u'seda_access_restriction_code', u'preferred',
                                  u'AR038')
            profile = testutils.setup_seda_profile(cnx)
            profile.cw_set(modification_date=YESTERDAY)
            cnx.commit()
            archive = profile.archives[0]
            # edit composite children
            archive.seda_name[0].cw_set(value=u'archive name')
            cnx.commit()
            self.assertMDNow(profile)
            self.resetMD(cnx, profile)
            # edit link from root to an entity which is not part of the container
            agent = testutils.organization_unit(
                cnx, u'marcel', archival_roles=['deposit'])
            profile.cw_set(seda_transferring_agent=agent)
            cnx.commit()
            self.assertMDNow(profile)
            self.resetMD(cnx, profile)
            # edit link from composite children to an entity which is not part of the container
            code = testutils.concept(cnx, 'AR038')
            archive.seda_access_restriction_code[0].cw_set(seda_access_restriction_code_value=code)
            cnx.commit()
            self.assertMDNow(profile)
            self.resetMD(cnx, profile)
            archive.seda_access_restriction_code[0].cw_set(seda_access_restriction_code_value=None)
            cnx.commit()
            self.assertMDNow(profile)
            self.resetMD(cnx, profile)
            # edit relation to a composite children
            arc = cnx.create_entity('SEDAAppraisalRuleCode')
            ard = cnx.create_entity('SEDAAppraisalRuleDuration')
            ar = cnx.create_entity('SEDAAppraisalRule', seda_appraisal_rule_code=arc,
                                   seda_appraisal_rule_duration=ard)
            archive.cw_set(seda_appraisal_rule=ar)
            cnx.commit()
            self.assertMDNow(profile)
            self.resetMD(cnx, profile)
            ar.cw_delete()
            cnx.commit()
            self.assertMDNow(profile)
            self.resetMD(cnx, profile)

    def test_externaluri_to_concept(self):
        with self.admin_access.repo_cnx() as cnx:
            # create some authority record and related objects
            agent = testutils.authority_record(cnx, u'bob')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            place = cnx.create_entity('AgentPlace', place_address=address, place_agent=agent)
            function = cnx.create_entity('AgentFunction', name=u'sponge', function_agent=agent)
            info = cnx.create_entity('LegalStatus', legal_status_agent=agent)
            cnx.commit()
            # create some external uri and link it to place, function and information entities
            exturi = cnx.create_entity('ExternalUri', cwuri=u'http://someuri/someobject',
                                       uri=u'http://someuri/someobject',
                                       reverse_equivalent_concept=[place, function, info])
            cnx.commit()
            # now insert a concept with the external uri as cwuri
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object', cwuri=u'http://someuri/someobject')
            cnx.commit()
            # ensure the external uri has been replaced by the concept and deleted
            place.cw_clear_all_caches()
            self.assertEqual(place.equivalent_concept[0].eid, concept.eid)
            function.cw_clear_all_caches()
            self.assertEqual(function.equivalent_concept[0].eid, concept.eid)
            info.cw_clear_all_caches()
            self.assertEqual(info.equivalent_concept[0].eid, concept.eid)
            self.failIf(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_externaluri_to_authorityrecord_subject(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            exturi = cnx.create_entity('ExternalUri', cwuri=u'a/b/c', uri=u'a/b/c')
            arelation = cnx.create_entity('AssociationRelation',
                                          association_from=bob,
                                          association_to=exturi)
            cnx.commit()
            alice = testutils.authority_record(cnx, u'alice', cwuri=u'a/b/c')
            cnx.commit()
            arelation.cw_clear_all_caches()
            self.assertEqual(arelation.association_to[0], alice)
            self.failIf(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_externaluri_to_authorityrecord_object(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            exturi = cnx.create_entity('ExternalUri', cwuri=u'a/b/c', uri=u'a/b/c')
            arelation = cnx.create_entity('AssociationRelation',
                                          association_from=exturi,
                                          association_to=bob)
            cnx.commit()
            alice = testutils.authority_record(cnx, u'alice', cwuri=u'a/b/c')
            cnx.commit()
            arelation.cw_clear_all_caches()
            self.assertEqual(arelation.association_from[0], alice)
            self.failIf(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_ark_generation_authorityrecord(self):
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'bob')
            cnx.commit()
            self.assertEqual(agent.ark, '0/r%09d' % agent.eid)
            self.assertEqual(agent.cwuri, 'ark:/0/r%09d' % agent.eid)
            agent = testutils.authority_record(cnx, u'john', ark=u'authority/123456')
            cnx.commit()
            self.assertEqual(agent.ark, 'authority/123456')
            self.assertEqual(agent.cwuri, 'ark:/authority/123456')
            agent = testutils.authority_record(cnx, u'alf', cwuri=u'http://someuri/someagent')
            cnx.commit()
            self.assertEqual(agent.ark, '0/r%09d' % agent.eid)
            self.assertEqual(agent.cwuri, 'http://someuri/someagent')

    def test_ark_generation_seda_profile(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            self.assertEqual(profile.ark, '0/p{0:09d}'.format(profile.eid))
            self.assertEqual(profile.cwuri, 'ark:/0/p{0:09d}'.format(profile.eid))

    def test_ark_generation_seda_profile_ark_given(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx, ark=u'authority/124')
            self.assertEqual(profile.ark, 'authority/124')
            self.assertEqual(profile.cwuri, 'ark:/authority/124')

    def test_ark_generation_seda_profile_with_cwuri(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(
                cnx, cwuri=u'http://example.org/profile/125')
            self.assertEqual(profile.ark, '0/p{0:09d}'.format(profile.eid))
            self.assertEqual(profile.cwuri, 'http://example.org/profile/125')

    def test_ark_generation_concept(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object')
            cnx.commit()
            self.assertEqual(scheme.ark, '0/v%09d' % scheme.eid)
            self.assertEqual(scheme.cwuri,
                             'ark:/0/v%09d' % scheme.eid)
            self.assertEqual(concept.ark, '0/c%09d' % concept.eid)
            self.assertEqual(concept.cwuri,
                             'ark:/0/c%09d' % concept.eid)
            scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://someuri/somescheme',
                                       ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object', cwuri=u'http://someuri/someconcept')
            cnx.commit()
            self.assertEqual(scheme.ark, '0/v%09d' % scheme.eid)
            self.assertEqual(scheme.cwuri, 'http://someuri/somescheme')
            self.assertEqual(concept.ark, '0/c%09d' % concept.eid)
            self.assertEqual(concept.cwuri, 'http://someuri/someconcept')
            scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://dcf/res/ark:/67717/Matiere',
                                       ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object', cwuri=u'http://dcf/res/ark:/67717/1234')
            cnx.commit()
            self.assertEqual(scheme.ark, '67717/Matiere')
            self.assertEqual(scheme.cwuri, 'http://dcf/res/ark:/67717/Matiere')
            self.assertEqual(concept.ark, '67717/1234')
            self.assertEqual(concept.cwuri, 'http://dcf/res/ark:/67717/1234')


class ExtractArkTC(unittest.TestCase):

    def test_ok(self):
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234'), '67717/1234')
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234#something'), '67717/1234')
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234?value'), '67717/1234')
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234/sub'), '67717/1234')
        self.assertEqual(extract_ark('ark:/67717/1234'), '67717/1234')

    def test_ko(self):
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717'), None)
        self.assertEqual(extract_ark('http://someuri/67717/1234'), None)


class EntityLifeCycleTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = self.create_user(cnx, u'bob',
                                   authority=cnx.find('Organization').one())
            bob_agent = testutils.agent(cnx, u'bob', agent_user=bob)
            cnx.commit()
            self.bob_eid = bob_agent.eid

    def _check_create(self, cnx, eid, msg):
        activity = cnx.find('Activity', type=u'create', used=eid).one()
        self.assertEqual(activity.associated_with[0].eid, self.bob_eid)
        self.assertEqual(activity.generated[0].eid, eid)
        self.assertEqual(activity.description, msg)

    def _check_modification(self, cnx, eid, msg):
        activity = cnx.find('Activity', type=u'modify', used=eid).one()
        self.assertEqual(activity.associated_with[0].eid, self.bob_eid)
        self.assertEqual(activity.used[0].eid, eid)
        self.assertEqual(activity.generated[0].eid, eid)
        self.assertEqual(activity.description, msg)

    def test_authorityrecord_create_update(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            cnx.commit()
            self._check_create(cnx, authorityrecord.eid,
                               'created authorityrecord')
            authorityrecord.cw_set(isni=u'123', name=u'Adam')
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'modified isni, name')

    def test_authorityrecord_add_component(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            cnx.commit()
            cnx.create_entity('AgentFunction', name=u'secret agent',
                              function_agent=authorityrecord)
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'added function_agent_object')

    def test_authorityrecord_delete_component(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            af = cnx.create_entity('AgentFunction', name=u'secret',
                                   function_agent=authorityrecord)
            cnx.commit()
            af.cw_delete()
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'removed function_agent_object')

    def test_authorityrecord_update_component(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            pn = cnx.create_entity('AgentFunction', name=u'secret agent',
                                   function_agent=authorityrecord)
            cnx.commit()
            pn.cw_set(name=u'boss of secret agents')
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'modified function_agent_object')

    def test_authorityrecord_add_subcomponent(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            cnx.commit()
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'added place_address, place_agent_object')

    def test_authorityrecord_delete_subcomponent(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            cnx.commit()
            address.cw_delete()
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid, 'removed place_address')

    def test_authorityrecord_update_subcomponent(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            cnx.commit()
            address.cw_set(street=u"1 avenue de l'Europe")
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'modified place_address')

    def test_authorityrecord_multi_modification(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            f1 = cnx.create_entity('AgentFunction', name=u'special agent',
                                   function_agent=authorityrecord)
            gc = cnx.create_entity('GeneralContext', content=u'sunglasses',
                                   general_context_of=authorityrecord)
            cnx.commit()
            gc.cw_set(content=u'stetson')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            f1.cw_delete()
            cnx.create_entity('AgentFunction', name=u'mib',
                              function_agent=authorityrecord)
            authorityrecord.cw_set(isni=u'123', name=u'Adam')
            cnx.commit()
            self._check_modification(
                cnx, authorityrecord.eid,
                '* modified function_agent_object, general_context_of_object, isni, name\n'
                '* added place_address, place_agent_object')

    def test_authorityrecord2authorityrecord_relation(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            k = testutils.authority_record(cnx, u'authorityrecord K')
            d = testutils.authority_record(cnx, u'authorityrecord D')
            cnx.commit()
            cnx.create_entity('AssociationRelation', association_from=k, association_to=d)
            cnx.commit()
            self._check_modification(cnx, k.eid, 'added association_from_object')
            self._check_modification(cnx, d.eid, 'added association_to_object')

    def test_no_activity_generated(self):
        with self.admin_access.repo_cnx() as cnx:
            agent = cnx.create_entity('Agent', name=u'smith',
                                      authority=cnx.find('Organization')[0][0])
            cnx.commit()
            authorityrecord = testutils.authority_record(
                cnx, u'Smith', reverse_authority_record=agent)
            cnx.commit()
            agent.cw_set(name=u'jones')
            cnx.commit()
            self.failIf(cnx.find('Activity', type=u'modify',
                                 used=authorityrecord.eid))

    def test_concept(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus',
                                       ark_naa=testutils.naa(cnx))
            cnx.commit()
            # concept scheme creation
            self._check_create(cnx, scheme.eid, 'created conceptscheme')
            concept = scheme.add_concept(u'hello')
            cnx.commit()
            # concept scheme add concept
            self._check_modification(cnx, scheme.eid, 'added in_scheme_object')
            # concept creation
            self._check_create(cnx, concept.eid, 'created concept')
            subconcept = concept.add_concept(u'goodbye')
            cnx.commit()
            # concept add concept
            self._check_modification(cnx, concept.eid, 'added broader_concept_object')
            # subconcept creation
            self._check_create(cnx, subconcept.eid, 'created concept')
            # subconcept creation scheme
            rset = cnx.execute('Activity X ORDERBY X DESC WHERE X type "modify", X used %(x)s',
                               {'x': scheme.eid})
            self.assertEqual(len(rset), 2)
            self.assertEqual(rset.get_entity(0, 0).description, 'added in_scheme_object')


class SEDAProfileHooksTC(CubicWebTC):

    def test_archive_created(self):
        """Test hook creating a Profile Archive upon creation of a SEDAProfile."""
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx, complete_archive=False)
            cnx.commit()
            arch = profile.archives[0]
            self.assertTrue(arch.seda_access_restriction_code)
            self.assertTrue(arch.seda_content_description)
            content_desc = arch.content_description
            self.assertTrue(content_desc.seda_description_level)
            self.assertEqual(profile.support_seda_exports, 'SEDA-0.2.xsd, SEDA-1.0.xsd')

    def test_profile_deprecated(self):
        """Test hook deprecating a SEDA Profile upon successor publication."""
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx, complete_archive=False)
            cnx.commit()
            workflow = profile.cw_adapt_to('IWorkflowable')
            workflow.fire_transition('publish')
            cnx.commit()
            profile.cw_clear_all_caches()
            self.assertEqual(workflow.state, 'published')
            cloned = cnx.create_entity('SEDAProfile', title=u'Clone', seda_replace=profile,
                                       ark_naa=testutils.naa(cnx))
            cnx.commit()
            cloned.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            profile.cw_clear_all_caches()
            self.assertEqual(workflow.state, 'deprecated')

    def test_profile_export_compat(self):
        with self.admin_access.client_cnx() as cnx:
            profile = cnx.create_entity('SEDAProfile', ark_naa=testutils.naa(cnx))
            cnx.commit()
            self.assertEqual(profile.support_seda_exports, '')
            ao = cnx.create_entity('ProfileArchiveObject',
                                   user_cardinality=u'0..1',
                                   seda_name=cnx.create_entity('SEDAName'),
                                   seda_parent=profile)
            cnx.commit()
            self.assertEqual(profile.support_seda_exports, '')
            ao.cw_set(user_cardinality=u'1')
            cnx.commit()
            profile.cw_clear_all_caches()
            self.assertEqual(profile.support_seda_exports, 'SEDA-0.2.xsd, SEDA-1.0.xsd')


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref schema"""

from yams.buildobjs import (EntityType, RelationDefinition, String, Date, Bytes, Int,
                            RichString, ComputedRelation, DEFAULT_RELPERMS)
from yams.constraints import UniqueConstraint, BoundaryConstraint, Attribute

from cubicweb.schema import (RO_ATTR_PERMS, ERQLExpression, RRQLExpression,
                             RQLConstraint, RQLVocabularyConstraint, WorkflowableEntityType,
                             make_workflowable)
from cubicweb.schemas.base import ExternalUri, EmailAddress

from cubes.skos.schema import ConceptScheme
from cubes.prov.schema import Agent

from cubes.saem_ref import (PERMISSIONS_GRAPHS, OUTER_RELATIVE_PERM_RELATIONS,
                            optional_relations, mandatory_rdefs)

_ = unicode


def dated_entity_type(cls):
    """Class decorator adding `start_date` and `end_date` attribute to an
    EntityType.
    """
    cls.add_relation(Date(constraints=[BoundaryConstraint(
        '<=', Attribute('end_date'), msg=_('start date must be less than end date'))]),
        name='start_date')
    cls.add_relation(Date(), name='end_date')
    return cls


def xml_wrap(cls):
    """Class decorator adding an `xml_wrap` attribute to an EntityType."""
    desc = _('XML elements not contained in EAC-CPF namespace')
    cls.add_relation(Bytes(description=desc), name='xml_wrap')
    return cls


def seda_profile_element(cardinalities=None, default_cardinality=None):
    """Class decorator adding attributes to configure a SEDA field.
    """
    def decorator(cls):
        if cardinalities:
            cls.add_relation(String(required=True, vocabulary=cardinalities,
                                    default=default_cardinality,
                                    internationalizable=True),
                             name='user_cardinality')
        cls.add_relation(String(fulltextindexed=True),
                         name='user_annotation')
        return cls
    return decorator


def publication_permissions(cls):
    """Set __permissions__ of `cls` entity type class preventing modification
    when not in state "draft".
    """
    cls.__permissions__ = cls.__permissions__.copy()
    cls.__permissions__['update'] = (
        ERQLExpression('U in_group G, G name IN ("managers", "users"), '
                       'X in_state ST, ST name "draft"'),
    )
    cls.__permissions__['delete'] = (
        ERQLExpression('U in_group G, G name IN ("managers", "users"), '
                       'X in_state ST, ST name "draft"'),
    )
    return cls


def groups_permissions(cls):
    """Set __permissions__ of `cls` entity type class preventing modification
    when user is not in managers or users group.
    """
    cls.__permissions__ = cls.__permissions__.copy()
    cls.__permissions__['update'] = (
        ERQLExpression('U in_group G, G name IN ("managers", "users")', 'U'),
    )
    return cls


def _rel_expr(rtype, role):
    return {'subject': 'X {rtype} A',
            'object': 'A {rtype} X'}[role].format(rtype=rtype)


def relative_update_permissions_through(*relations, **options):
    """Class decorator setting 'update' permission on entity type class relative to
    `(rtype, role)` relations.

    This is for use on entities whose parent is reachable through non mandatory relations, others
    should be handled by :func:`graph_set_etypes_update_permissions`.
    """
    handle_not_exists = options.pop('may_not_exists', True)
    assert not options, 'unknown options: %s' % options

    def set_perms(cls):
        cls.__permissions__ = cls.__permissions__.copy()
        exprs = []
        for rtype, role in relations:
            relexpr = _rel_expr(rtype, role)
            if handle_not_exists:
                exprs.append('NOT EXISTS({relexpr}), U in_group G, '
                             'G name IN ("managers", "users")'.format(relexpr=relexpr))
            exprs.append('{relexpr}, U has_update_permission A'.format(relexpr=relexpr))
        for action in ('update', 'delete'):
            cls.__permissions__[action] = tuple(ERQLExpression(e) for e in exprs)
        return cls
    return set_perms


def graph_set_etypes_update_permissions(schema, graph, etype):
    """Add `action` permissions for all entity types in the composite `graph`
    with root `etype`. Respective permissions that are inserted on each
    entity type are relative to the "parent" in the relation from this
    entity type walking up to the graph root.

    So for instance, calling `set_etype_permissions('R', 'update')`
    on a schema where `A related_to B` and `R root_of B` one will get:

    * "U has_update_permission R, R root_of X" for `B` entity type and,
    * "U has_update_permission P, X related_to P" for `A` entity type.

    If an entity type in the graph is reachable through multiple relations, a
    permission for each of this relation will be inserted so that if any of
    these match, the permission check will succeed.
    """
    structure = graph.parent_structure(etype)
    optionals = optional_relations(schema, structure)
    for child, relations in structure.iteritems():
        skiprels = optionals.get(child, set())
        exprs = []
        for rtype, role in relations:
            if (rtype, role) in skiprels:
                continue
            relexpr = _rel_expr(rtype, role)
            exprs.append('{relexpr}, U has_update_permission A'.format(relexpr=relexpr))
        if exprs:
            for action in ('update', 'delete'):
                schema[child].set_action_permissions(action,
                                                     tuple(ERQLExpression(e) for e in exprs))


_RELATIVE_PERM_RELATIONS_CONSISTENCY_CHECK = set()


def relative_write_permissions_of(role):
    """Relation definition class decorator setting permissions for 'add' and 'delete' actions
    related to `role` target.

    Important notes:

    - this is for use with non mandatory relations or relation to some entity which doesn't belong
      to the compound graph, others should be handled by :func:`graph_set_write_rdefs_permissions`,

    - the decorated relation definition must be added to `cubicweb.server.ON_COMMIT_ADD_RELATIONS`.
    """
    def set_perms(cls):
        _RELATIVE_PERM_RELATIONS_CONSISTENCY_CHECK.add(getattr(cls, 'name', cls.__name__))
        if not cls.__permissions__:
            cls.__permissions__ = DEFAULT_RELPERMS
        cls.__permissions__ = cls.__permissions__.copy()
        var = {'object': 'O', 'subject': 'S'}[role]
        expr = 'U has_update_permission {0}'.format(var)
        for action in ('add', 'delete'):
            cls.__permissions__[action] = (RRQLExpression(expr),)
        return cls
    return set_perms


def graph_set_write_rdefs_permissions(schema, graph, etype):
    """Set 'add' and 'delete' permissions for all mandatory relation definitions in the composite
    `graph` with root `etype`.

    Respective permissions that are inserted on each relation definition are relative to the
    "parent" in the relation from this entity type walking up to the graph root.

    Relations which are not mandatory or which are not part of the graph structure should be handled
    manually.
    """
    structure = graph.parent_structure(etype)
    for rdef, parent_role in mandatory_rdefs(schema, structure):
        var = {'object': 'O', 'subject': 'S'}[parent_role]
        expr = 'U has_update_permission {0}'.format(var)
        for action in ('add', 'delete'):
            rdef.set_action_permissions(action, (RRQLExpression(expr), ))


# Disable "update" for ExternalUri as these should only come from imported data
# and are meant to only be created or deleted.
ExternalUri.__permissions__ = ExternalUri.__permissions__.copy()
ExternalUri.__permissions__['update'] = ()


# Customization of EmailAddress entity type.
EmailAddress.remove_relation('alias')


# Customization of Agent entity type.
make_workflowable(Agent)
groups_permissions(Agent)
Agent.add_relation(String(required=True, fulltextindexed=True), name='name')


@dated_entity_type
@groups_permissions
class AuthorityRecord(WorkflowableEntityType):
    name = String(required=True, fulltextindexed=True)
    isni = String(unique=True,
                  description=_('International Standard Name Identifier'))

# Customization of skos schema.
make_workflowable(ConceptScheme)
publication_permissions(ConceptScheme)


class Organization(EntityType):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'update': ('managers', ),
        'delete': ('managers', ),
    }
    name = String(required=True, fulltextindexed=True, unique=True)


class OrganizationUnit(WorkflowableEntityType):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'update': ('managers', ),
        'delete': ('managers', ),
    }
    __unique_together__ = [('name', 'authority')]
    name = String(required=True, fulltextindexed=True, unique=True)


class user_authority(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers',),
        'delete': (),
    }
    name = 'authority'
    subject = 'CWUser'
    object = 'Organization'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint('NOT EXISTS(A agent_user S) '
                      'OR EXISTS(B agent_user S, B authority O)'),
    ]


class others_authority(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', RRQLExpression('U authority O'),),
        'delete': (),
    }
    name = 'authority'
    subject = ('OrganizationUnit', 'Agent')
    object = 'Organization'
    cardinality = '1*'
    composite = 'object'
    inlined = True


class agent_user(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers',),
        'delete': ('managers',),
    }
    subject = 'Agent'
    object = 'CWUser'
    cardinality = '??'
    inlined = True
    description = _('the application user related to this agent')
    constraints = [
        RQLConstraint('NOT EXISTS(O authority A) '
                      'OR EXISTS(O authority B, S authority B)'),
    ]


class _authority_record(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers',),
        'delete': ('managers',),
    }
    name = 'authority_record'
    object = 'AuthorityRecord'
    cardinality = '??'
    inlined = True
    description = _('the authority record describing this agent')


class ou_authority_record(_authority_record):
    subject = 'OrganizationUnit'
    description = _('the authority record describing this organization unit')
    constraints = [
        RQLConstraint('O agent_kind K, K name "authority"'),
    ]


class agent_authority_record(_authority_record):
    subject = 'Agent'
    description = _('the authority record describing this agent')
    constraints = [
        RQLConstraint('O agent_kind K, K name "person"'),
    ]


class contact_point(RelationDefinition):
    subject = 'OrganizationUnit'
    object = 'Agent'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint('S authority A, O authority A'),
        RQLVocabularyConstraint('O in_state ST, ST name "published"'),
    ]
    description = _('set an agent as the contact point of an organization unit')


class archival_unit(RelationDefinition):
    subject = 'Organization'
    object = 'OrganizationUnit'
    cardinality = '?*'
    description = _("the archival unit responsible for dealing with the organization's "
                    "documents")
    constraints = [RQLConstraint('O archival_role AR, AR name "archival"'),
                   RQLVocabularyConstraint('O in_state ST, ST name "published"')]


class ArchivalRole(EntityType):
    """An archival role determines the kind of action (e.g. deposit or control)
    an agent may perform on an archive entity.
    """
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'update': ('managers', ),
        'delete': ('managers', ),
    }
    name = String(required=True, unique=True, internationalizable=True)


class archival_role(RelationDefinition):
    subject = 'OrganizationUnit'
    object = 'ArchivalRole'
    cardinality = '**'
    description = _("the organization unit's archival role (producer, control, etc.)")


class use_email(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests',),
        'add': ('managers', RRQLExpression('U has_update_permission S'),),
        'delete': ('managers', RRQLExpression('U has_update_permission S'),),
    }
    subject = 'Agent'
    object = 'EmailAddress'
    cardinality = '*?'
    composite = 'subject'
    fulltext_container = 'subject'


class phone_number(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests',),
        'add': ('managers', RRQLExpression('U has_update_permission S'),),
        'delete': ('managers', RRQLExpression('U has_update_permission S'),),
    }
    subject = 'Agent'
    object = 'PhoneNumber'
    cardinality = '*1'
    composite = 'subject'


# EAC-CPF model elements for AuthorityRecord #############################################

class AgentFunction(EntityType):
    """The function of an AuthorityRecord"""
    name = String(fulltextindexed=True, internationalizable=True)
    description = RichString(fulltextindexed=True)
    __unique_together__ = [('name', 'function_agent')]


class function_agent(RelationDefinition):
    subject = 'AgentFunction'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True


class AgentPlace(EntityType):
    """Qualified relation between an AuthorityRecord and a PostalAddress"""
    name = String(fulltextindexed=True,
                  description=_('encoded information about the address (e.g. '
                                '"Paris, France")'))
    role = String(description=_('contextual role the address has in relation '
                                'with the agent (e.g. "home")'),
                  internationalizable=True)
    __unique_together__ = [('role', 'place_agent', 'place_address')]


class place_address(RelationDefinition):
    subject = 'AgentPlace'
    object = 'PostalAddress'
    cardinality = '?1'
    inlined = True
    composite = 'subject'
    fulltext_container = 'subject'


class place_agent(RelationDefinition):
    subject = 'AgentPlace'
    object = 'AuthorityRecord'
    cardinality = '1*'
    inlined = True
    composite = 'object'
    fulltext_container = 'object'


class postal_address(ComputedRelation):
    rule = 'P place_agent S, P place_address O'


class AgentKind(EntityType):
    """A kind of agent (e.g. "person", "authority" or "family")"""
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'update': (),
        'delete': (),
    }
    name = String(required=True, unique=True, internationalizable=True)


class agent_kind(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'users'),
        'delete': (RRQLExpression('O name "unknown-agent-kind"'),),
    }
    subject = 'AuthorityRecord'
    object = 'AgentKind'
    cardinality = '1*'
    inlined = True


class related_concept_scheme(RelationDefinition):
    subject = 'OrganizationUnit'
    object = 'ConceptScheme'
    cardinality = '**'
    description = _('concept schemes used by the agent')


class _agent_relation(RelationDefinition):
    """Abstract relation between agents"""
    subject = None
    object = ('AuthorityRecord', 'ExternalUri')
    cardinality = '1*'
    inlined = True


class GeneralContext(EntityType):
    """Information about the general social and cultural context of an agent"""
    content = RichString(fulltextindexed=True)


class general_context_of(RelationDefinition):
    subject = 'GeneralContext'
    object = 'AuthorityRecord'
    cardinality = '1*'
    inlined = True
    composite = 'object'
    fulltext_container = 'object'


@xml_wrap
@dated_entity_type
class AssociationRelation(EntityType):
    """Association relation between agents"""
    description = RichString()


class association_from(_agent_relation):
    subject = 'AssociationRelation'


class association_to(_agent_relation):
    subject = 'AssociationRelation'


@xml_wrap
class ChronologicalRelation(EntityType):
    """Chronological relation between agents"""
    description = RichString()


class chronological_predecessor(_agent_relation):
    subject = 'ChronologicalRelation'


class chronological_successor(_agent_relation):
    subject = 'ChronologicalRelation'


@xml_wrap
@dated_entity_type
class HierarchicalRelation(EntityType):
    """Hierarchical relation between agents"""
    description = RichString()


class hierarchical_parent(_agent_relation):
    subject = 'HierarchicalRelation'


class hierarchical_child(_agent_relation):
    subject = 'HierarchicalRelation'


class generated(RelationDefinition):
    subject = 'Activity'
    object = ('AuthorityRecord', 'Concept', 'ConceptScheme')


class used(RelationDefinition):
    subject = 'Activity'
    object = ('AuthorityRecord', 'Concept', 'ConceptScheme')


@dated_entity_type
class Mandate(EntityType):
    """Reference text coming from an authority"""
    term = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


@dated_entity_type
class LegalStatus(EntityType):
    """Information relative to the legal status of an authority"""
    term = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


class History(EntityType):
    """Biographical or historical information"""
    text = RichString(fulltextindexed=True)


class Structure(EntityType):
    """Information about the structure of an authority"""
    description = RichString(fulltextindexed=True)


@dated_entity_type
class Occupation(EntityType):
    term = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


class occupation_agent(RelationDefinition):
    subject = 'Occupation'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True
    description = _('occupation in which the person works or has worked')


class mandate_agent(RelationDefinition):
    subject = 'Mandate'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True
    description = _('mandate of an AuthorityRecord')


class legal_status_agent(RelationDefinition):
    subject = 'LegalStatus'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True
    description = _('legal status of an AuthorityRecord')


class structure_agent(RelationDefinition):
    subject = 'Structure'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True
    description = _('information about the structure of an AuthorityRecord')


class history_agent(RelationDefinition):
    subject = 'History'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True
    description = _('information about the history of an AuthorityRecord')


class Citation(EntityType):
    note = RichString()
    uri = String()


class has_citation(RelationDefinition):
    subject = ('GeneralContext', 'Mandate', 'Occupation', 'History',
               'AgentFunction', 'LegalStatus', 'AgentPlace')
    object = 'Citation'
    cardinality = '*?'
    composite = 'subject'
    description = _('reference to an external citation resource')


@xml_wrap
@dated_entity_type
class EACResourceRelation(EntityType):
    """Represent a relation between an AuthorityRecord and a remote resource in the
    EAC-CPF model.
    """
    agent_role = String(description=_('type of relation the agent has to the resource'),
                        internationalizable=True)
    resource_role = String(description=_('type or nature of the remote resource'),
                           internationalizable=True)
    description = RichString(fulltextindexed=True)


class resource_relation_agent(RelationDefinition):
    subject = 'EACResourceRelation'
    object = 'AuthorityRecord'
    cardinality = '1*'
    inlined = True
    composite = 'object'
    fulltext_container = 'object'


class resource_relation_resource(RelationDefinition):
    subject = 'EACResourceRelation'
    object = 'ExternalUri'
    cardinality = '1*'
    inlined = True


@xml_wrap
class EACSource(EntityType):
    """A source used to establish the description of an AuthorityRecord"""
    title = String(fulltextindexed=True)
    url = String()
    description = RichString(fulltextindexed=True)


class source_agent(RelationDefinition):
    subject = 'EACSource'
    object = 'AuthorityRecord'
    cardinality = '1*'
    composite = 'object'
    fulltext_container = 'object'
    inlined = True


class vocabulary_source(RelationDefinition):
    subject = ('Mandate', 'LegalStatus', 'AgentFunction', 'AgentPlace',
               'Occupation')
    object = 'ConceptScheme'
    cardinality = '?*'


class equivalent_concept(RelationDefinition):
    subject = ('Mandate', 'LegalStatus', 'AgentFunction', 'AgentPlace',
               'Occupation')
    object = ('ExternalUri', 'Concept')
    constraints = [RQLVocabularyConstraint('S vocabulary_source SC, O in_scheme SC')]
    cardinality = '?*'
    # relation with 'ExternalUri' as object can't be inlined because of a limitation of
    # data-import's (massive) store
    inlined = False


# ARK ##########################################################################

class ark(RelationDefinition):
    __permissions__ = RO_ATTR_PERMS
    subject = (
        'Agent',
        'AuthorityRecord',
        'Concept',
        'ConceptScheme',
        'OrganizationUnit',
        'SEDAProfile',
    )
    object = 'String'
    description = _('ARK Identifier - will be generated if not specified')
    constraints = [UniqueConstraint()]
    cardinality = '11'


class ArkNameAssigningAuthority(EntityType):
    """Name Assigning Authority (NAA) for ARK generation."""
    who = String(required=True, unique=True,
                 description=_('official organization name'))
    what = Int(required=True, unique=True,
               description=_('Name Assigning Authority Number (NAAN)'))


class _ark_naa(RelationDefinition):
    name = 'ark_naa'
    object = 'ArkNameAssigningAuthority'
    inlined = True


class organization_ark_naa(_ark_naa):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'delete': ('managers', ),
    }
    subject = 'Organization'
    cardinality = '?*'
    description = _("ARK identifier Name Assigning Authority (NAA) - "
                    "you'll need one to start creating objects")


class mandatory_ark_naa(_ark_naa):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'users',),
        'delete': (),
    }
    subject = ('AuthorityRecord', 'SEDAProfile')
    cardinality = '1*'
    description = _("ARK identifier Name Assigning Authority (NAA)")


class optional_ark_naa(_ark_naa):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'users',),
        'delete': (),
    }
    subject = ('ConceptScheme', 'SKOSSource')
    cardinality = '?*'
    description = _("ARK identifier Name Assigning Authority (NAA), "
                    "necessary to create ARK for concepts which don't have one yet.")


# SEDA #######################################################################

@publication_permissions
class SEDAProfile(WorkflowableEntityType):
    """a SEDA profile allows to restrict the `SEDA schema`_ by generating an XML schema (XSD) that
    may be transferred to SEDA compliant platforms (eg Alfresco and As@lae).

    .. _`SEDA schema`:http://www.archivesdefrance.culture.gouv.fr/seda/api/index.html
    """
    title = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)
    support_seda_exports = String(__permissions__={'read': ('managers', 'users', 'guests',),
                                                   'add': (),
                                                   'update': ()})


class seda_transferring_agent(RelationDefinition):  # XXX OrganizationUnit use_profile Profile
    subject = 'SEDAProfile'
    object = 'OrganizationUnit'
    cardinality = '**'
    constraints = [RQLConstraint('O archival_role R, R name "deposit"')]


@relative_write_permissions_of('object')
class _seda_parent(RelationDefinition):
    name = 'seda_parent'
    composite = fulltext_container = 'object'
    inlined = True


class seda_parent_attachement(_seda_parent):
    subject = 'ProfileDocument'
    object = 'ProfileArchiveObject'
    cardinality = '?*'


class seda_parent_documentunit(_seda_parent):
    subject = 'ProfileArchiveObject'
    object = ('ProfileArchiveObject', 'SEDAProfile')
    cardinality = '?*'


@relative_update_permissions_through(('seda_parent', 'subject'))
@seda_profile_element(cardinalities=['0..1', '0..n', '1', '1..n'], default_cardinality='1..n')
class ProfileArchiveObject(EntityType):
    """Archive object of a SEDA profile. There may be zero to n archive object in a profile archive
    or archive object.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Element_ArchiveObject.htm#ArchiveObject

    Merged with Archive of a SEDA profile. There may be one to n archives in a profile.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Element_Archive.htm#Archive
    """


@relative_update_permissions_through(('seda_parent', 'subject'))
@seda_profile_element(cardinalities=['0..1', '0..n', '1', '1..n'], default_cardinality='0..n')
class ProfileDocument(EntityType):
    """Document of a SEDA profile. There may be zero to n documents in a profile archive or archive
    object.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Element_Document.htm#Document
    """


class seda_file_type_code(RelationDefinition):
    subject = 'ProfileDocument'
    object = 'SEDAFileTypeCode'
    cardinality = '?1'
    composite = fulltext_container = 'subject'
    inlined = True


class seda_character_set_code(RelationDefinition):
    subject = 'ProfileDocument'
    object = 'SEDACharacterSetCode'
    cardinality = '?1'
    composite = fulltext_container = 'subject'
    inlined = True


class seda_document_type_code(RelationDefinition):
    subject = 'ProfileDocument'
    object = 'SEDADocumentTypeCode'
    cardinality = '11'
    composite = fulltext_container = 'subject'
    inlined = True


@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='0..1')
class SEDAFileTypeCode(EntityType):
    """File type code of a document profile

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_filetype_code_xsd_Element_FileTypeCode.htm#FileTypeCode
    """


@relative_write_permissions_of('subject')
class seda_file_type_code_value(RelationDefinition):
    subject = 'SEDAFileTypeCode'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_file_type_code", CR is CWRType')]


@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='0..1')
class SEDACharacterSetCode(EntityType):
    """Character set code of a document profile

    http://www.archivesdefrance.culture.gouv.fr/seda/api/UNECE_CharacterSetEncodingCode_40106_xsd_Simple_Type_clm60133_CharacterSetEncodingCodeContentType.htm#CharacterSetEncodingCodeContentType
    """


@relative_write_permissions_of('subject')
class seda_character_set_code_value(RelationDefinition):
    subject = 'SEDACharacterSetCode'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_character_set_code", CR is CWRType')]


@seda_profile_element()
class SEDADocumentTypeCode(EntityType):
    """Document type code for a SEDA Document Profile

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_documenttype_code_xsd_Simple_Type_DocumentTypeCodeType.htm#DocumentTypeCodeType
    """


@relative_write_permissions_of('subject')
class seda_document_type_code_value(RelationDefinition):
    subject = 'SEDADocumentTypeCode'
    object = 'Concept'
    cardinality = '1*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_document_type_code", CR is CWRType')]


@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='0..1')
class SEDAAppraisalRule(EntityType):
    u"""Appraisal rule (règle de sort final) of a SEDA profile.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Element_AppraisalRule.htm#AppraisalRule
    """


@seda_profile_element()
class SEDAAppraisalRuleDuration(EntityType):
    u"""Duration of a SEDAAppraisalRule

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Complex_Type_AppraisalRuleType.htm#AppraisalRuleType_Duration
    """


@seda_profile_element()
class SEDAAppraisalRuleCode(EntityType):
    u"""Code of a SEDAAppraisalRule ('conserver', 'detruire')

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Complex_Type_AppraisalRuleType.htm#AppraisalRuleType_Code
    """


class seda_appraisal_rule(RelationDefinition):
    subject = 'ProfileArchiveObject'
    object = 'SEDAAppraisalRule'
    cardinality = '?1'
    composite = fulltext_container = 'subject'
    inlined = True


class seda_appraisal_rule_duration(RelationDefinition):
    subject = 'SEDAAppraisalRule'
    object = 'SEDAAppraisalRuleDuration'
    cardinality = '11'
    composite = fulltext_container = 'subject'
    inlined = True


class seda_appraisal_rule_duration_value(RelationDefinition):
    subject = 'SEDAAppraisalRuleDuration'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_appraisal_rule_duration", CR is CWRType')]


class seda_appraisal_rule_code(RelationDefinition):
    subject = 'SEDAAppraisalRule'
    object = 'SEDAAppraisalRuleCode'
    cardinality = '11'
    composite = fulltext_container = 'subject'
    inlined = True


@relative_write_permissions_of('subject')
class seda_appraisal_rule_code_value(RelationDefinition):
    subject = 'SEDAAppraisalRuleCode'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_appraisal_rule_code", CR is CWRType')]


# XXX SEDA 1.0 stricto-sensu would be
#   ProfilArchive-[0..1]->AccessRestrictionRule-1->AccessRestrictionCode
#                                              -1->DateDepartCalcul
#
# but we skip AccessRestrictionRule entity for now as DateDepartCalcul isn't handled yet.
#
# Regarding cardinalities: AccessRestrictionRule is mandatory from archive but optional from archive
# object. This is problematic since the user cardinality is stored on the target entity.  To handle
# this while keeping this modelization, we allow on the target entity the most relaxed cardinalities
# (['0..1', '1']) and will handle restriction for the archive case in UI/hooks.
@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='1')
class SEDAAccessRestrictionCode(EntityType):
    u"""Access resstriction code for an archive of a SEDA profile. There may be zero to one access
    control rule on an archive object but one and only one on an archive.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_accessrestriction_code_xsd_Element_AccessRestrictionCode.htm#AccessRestrictionCode
    """


class seda_access_restriction_code(RelationDefinition):
    subject = 'ProfileArchiveObject'
    object = 'SEDAAccessRestrictionCode'
    cardinality = '?1'
    composite = fulltext_container = 'subject'
    inlined = True


@relative_write_permissions_of('subject')
class seda_access_restriction_code_value(RelationDefinition):
    subject = 'SEDAAccessRestrictionCode'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_access_restriction_code", CR is CWRType')]


@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='1')
class SEDAContentDescription(EntityType):
    u"""Content description for an archive or archive object of a SEDA profile. There may be zero to
    one content description on an archive object by one and only one on an archive.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Element_ContentDescription.htm#ContentDescription
    """


class seda_content_description(RelationDefinition):
    subject = 'ProfileArchiveObject'
    object = 'SEDAContentDescription'
    cardinality = '?1'
    composite = fulltext_container = 'subject'
    inlined = True


@seda_profile_element()
class SEDADescriptionLevel(EntityType):
    """Description level for a content description of a SEDA profile.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Complex_Type_ContentDescriptionType.htm#ContentDescriptionType_DescriptionLevel
    """


class seda_description_level(RelationDefinition):
    subject = 'SEDAContentDescription'
    object = 'SEDADescriptionLevel'
    cardinality = '11'
    composite = fulltext_container = 'subject'
    inlined = True


@relative_write_permissions_of('subject')
class seda_description_level_value(RelationDefinition):
    subject = 'SEDADescriptionLevel'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [RQLConstraint('O in_scheme CS, CS scheme_relation CR, CR name '
                                 '"seda_description_level", CR is CWRType')]


@seda_profile_element(cardinalities=['0..1', '0..n', '1', '1..n'], default_cardinality='0..n')
class SEDAKeyword(EntityType):
    """Keyword for a content description of a SEDA profile.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Complex_Type_ContentDescriptionType.htm#ContentDescriptionType_Keyword
    """


class seda_keyword_of(RelationDefinition):
    subject = 'SEDAKeyword'
    object = 'SEDAContentDescription'
    cardinality = '1*'
    composite = fulltext_container = 'object'
    inlined = True


@relative_write_permissions_of('subject')
class seda_keyword_scheme(RelationDefinition):
    subject = 'SEDAKeyword'
    object = 'ConceptScheme'
    cardinality = '1*'
    inlined = True


@relative_write_permissions_of('subject')
class seda_keyword_value(RelationDefinition):
    subject = 'SEDAKeyword'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint('S seda_keyword_scheme KS, O in_scheme KS',
                      msg=_("concept doesn't belong to the scheme associated to the keyword")),
    ]


@relative_update_permissions_through(('seda_oldest_date', 'object'),
                                     ('seda_latest_date', 'object'),
                                     may_not_exists=False)
@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='0..1')
class SEDADate(EntityType):
    """Date used in a SEDA profile."""
    value = Date()


@relative_write_permissions_of('subject')
class seda_oldest_date(RelationDefinition):
    subject = 'SEDAContentDescription'
    object = 'SEDADate'
    cardinality = '??'
    composite = fulltext_container = 'subject'
    inlined = True
    constraints = [
        RQLConstraint('NOT EXISTS(D seda_latest_date O)',
                      msg=_('SEDA date cannot be used as oldest and latest')),
    ]


@relative_write_permissions_of('subject')
class seda_latest_date(RelationDefinition):
    subject = 'SEDAContentDescription'
    object = 'SEDADate'
    cardinality = '??'
    composite = fulltext_container = 'subject'
    inlined = True
    constraints = [
        RQLConstraint('NOT EXISTS(D seda_oldest_date O)',
                      msg=_('SEDA date cannot be used as oldest and latest')),
    ]


@seda_profile_element()
class SEDAName(EntityType):
    """Name of an archive of a SEDA profile or profile object.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Complex_Type_ArchiveType.htm#ArchiveType_Name
    """
    value = String(fulltextindexed=True)


class seda_name(RelationDefinition):
    subject = 'ProfileArchiveObject'
    object = 'SEDAName'
    cardinality = '11'
    composite = fulltext_container = 'subject'
    inlined = True


@seda_profile_element(cardinalities=['0..1', '1'], default_cardinality='0..1')
class SEDADescription(EntityType):
    """Description of a document of a SEDA profile. There may be zero to one description on a
    profile document.

    http://www.archivesdefrance.culture.gouv.fr/seda/api/seda_v1-0_archive_xsd_Complex_Type_DocumentType.htm#DocumentType_Description
    """
    value = String(fulltextindexed=True)  # XXX Rich String ?


class seda_description(RelationDefinition):
    subject = ('ProfileDocument', 'SEDAContentDescription')
    object = 'SEDADescription'
    cardinality = '?1'
    composite = fulltext_container = 'subject'
    inlined = True


class _seda_clone_of(RelationDefinition):  # XXX Profile clone_of Profile
    cardinality = '?*'
    inlined = True


class seda_clone_of_document(_seda_clone_of):
    name = 'seda_clone_of'
    subject = 'ProfileDocument'
    object = 'ProfileDocument'


class seda_clone_of_documentunit(_seda_clone_of):
    name = 'seda_clone_of'
    subject = 'ProfileArchiveObject'
    object = 'ProfileArchiveObject'


class scheme_relation(RelationDefinition):
    """Special relation from a concept scheme to a relation type, that may be used to restrict
    possible concept of a particular relation without depending on the scheme's name or other weak
    mecanism. See :class:`seda_access_restriction_code_value` for instance.
    """
    __permissions__ = {'read': ('managers', 'users', 'guests'),
                       'add': ('managers',),
                       'delete': ('managers',)}
    subject = 'ConceptScheme'
    object = 'CWRType'
    cardinality = '*?'


class seda_replace(RelationDefinition):  # XXX Profile new_version_of Profile
    __permissions__ = {'read': ('managers', 'users', 'guests'),
                       'add': ('managers', 'users',
                               RRQLExpression('O in_state ST, ST name "published"')),
                       'delete': ()}
    subject = 'SEDAProfile'
    object = 'SEDAProfile'
    cardinality = '??'
    inlined = True


def post_build_callback(schema):
    for etype, graph in PERMISSIONS_GRAPHS.iteritems():
        graph_set_etypes_update_permissions(schema, graph(schema), etype)
        graph_set_write_rdefs_permissions(schema, graph(schema), etype)
    assert _RELATIVE_PERM_RELATIONS_CONSISTENCY_CHECK == OUTER_RELATIVE_PERM_RELATIONS, \
        ('inconsistency between calls to relative_permission_of and '
         'OUTER_RELATIVE_PERM_RELATIONS: %s' % _RELATIVE_PERM_RELATIONS_CONSISTENCY_CHECK)
    # permissions override
    schema['Label'].set_action_permissions('delete', ('managers', 'users'))
    for rtype in ('in_scheme', 'broader_concept', 'label_of'):
        for rdef in schema[rtype].rdefs.values():
            rdef.set_action_permissions('add', ('managers', 'users'))
            if rtype == 'label_of':
                rdef.set_action_permissions('delete', ('managers', 'users'))
